#!/usr/bin/python

import string, types,sys, time
#import time,hexdump,pdb
ORBITLENGTH=3564
ErrorMessage=''
symname= ("PH", "PP", "CAL", "TOF")
def PrintError(fstr):
  global ErrorMessage
  print "Error:",fstr
  ErrorMessage= ErrorMessage+fstr+'\n'

def bit_not(n, numbits=32):
    return (1 << numbits) - 1 - n

class ST:
  RIGHT=1 # bracket
  LEFT=2
  BIT=3   # bit (H or L -value 1 or 0)
  REP=4   # repetitions (int)
  EOM=5   # End Of Mask
  def __init__(self, text):
    self.type=None
    self.value=None
    self.length=0
    if text=='': 
      self.type=ST.EOM; self.length=0
      #print "ST:EOM"
      return
    while 1:
      if text[self.length]==' ': 
        self.length=self.length+1
        if self.length==len(text):
          self.type=ST.EOM; return
      else: break
    if text[self.length]=='L' or text[self.length]=='l':
      self.type=ST.BIT; self.value=0; self.length=self.length+1
    elif text[self.length]=='H' or text[self.length]=='h':
      self.type=ST.BIT; self.value=1; self.length=self.length+1
    elif text[self.length]=='(':
      self.type=ST.LEFT;  self.length=self.length+1
    elif text[self.length]==')':
      self.type=ST.RIGHT;  self.length=self.length+1
    elif text[self.length] in string.digits:
      self.length=self.length+1
      while 1:
        #print "text:",text, self.length
        if self.length>(len(text)-1): 
          PrintError(text+" -premature end of mask definition (,L or H expected")
          break
        if text[self.length] in string.digits:
          self.length=self.length+1
        else: break
      self.type= ST.REP 
      self.value= int(text[:self.length])
    else:
      PrintError(" bad string:"+ text)
    #print "ST:", self.type, self.length, "val=",self.value

class BCmask:
  def __init__(self,board,bitname):
    """
    bitname: a bit position in bcmask array
    bcmexp: "25L 25(2H2LH 3(23HL))"
          - H/h -> 1  L/l -> 0
          - spaces, new lines are white characters
          - ignore comment lines starting with #
          - max. 3564 bits (warning for less or more)
          : None -> read the content from a board
    Operation: 
    - read all masks into self.bcmasks
    - self.bcmstr contains definition of a mask
    output: self.bcmstr initialised from a board or bcmexp
    """
    self.board= board
    self.bitname= bitname
    self.level=0
    #read from a board
    #self.bcmasks= self.board.ReadBlock("bcmasks.data", ORBITLENGTH, 0)
    # 
    msks_ValVector_uint32= self.board.ReadBlock("bcmasks.data", ORBITLENGTH, 0)
    msks= []
    for ix in range(len(msks_ValVector_uint32)):
      msks.append(msks_ValVector_uint32[ix])
    self.bcmasks= msks
    self.updatebcmstr()
  def prt(self):
    enabit= self.board.Read("ctpemu.BCmask_ena") & (1<<self.bitname)
    enadis= "enabled"
    if enabit==0: enadis= "disabled"
    print "%s %s, BCmask:"%(symname[self.bitname], enadis), self.bcmstr
  def updatebcmstr(self):
    """ bcmstr <- slf.bcmasks[][self.bitname]
    """
    longbmstr= ""
    for bc in range(ORBITLENGTH):
      if self.bcmasks[bc] & (1<<self.bitname):
        longbmstr= longbmstr+"H"
      else:
        longbmstr= longbmstr+"L"
    self.bcmstr= self.shrink(longbmstr)
  def updatehw(self, bcmexp):
    if bcmexp != None:
      ErrorMessage=''
      self.bcmstr= bcmexp
      print "old bcmstr:",self.bcmstr
      if (bcmexp.find("H")>=0) or (bcmexp.find("L")>=0):
        # 3560H4L
        bcmbits,chars= self.bcm2bits(0)
      else:
        bcmbits= []
        for bc in range(ORBITLENGTH):
          bcmbits.append(0)
        expints= string.split(bcmexp)
        processed= ""
        for bcn in expints:
          try:
            bcmbits[int(bcn)]= 1
          except:
            print "bad int:", bcn
            continue
          if processed != "": processed= processed+" "
          processed= processed + bcn
        chars= len(processed)
      # bcmbits, chars filled now. ErrorMessage valid if chars != len(bcmbits)
      if len(bcmbits)>ORBITLENGTH:
        print "Error: Too long BCmask definition (%d BCs)"%len(bcmbits)
        del(bcmbits[ORBITLENGTH:])
      elif len(bcmbits)<ORBITLENGTH:
        print "Warning: short BCmask definition (%d BCs)"%len(bcmbits)
      else:
        pass
      if chars != len(self.bcmstr):
        print ErrorMessage
      for bc in range(len(bcmbits)):
        if bcmbits[bc]==1:
          self.bcmasks[bc]= self.bcmasks[bc] | (1<<self.bitname)
        else:
          self.bcmasks[bc]= self.bcmasks[bc] & (~(1<<self.bitname))
      # recreate bcmstr from self.bcmasks[][self.bitname] -it will reflect
      # the board even in case of errorneous bcmexp
      self.updatebcmstr()
      print "updating hw with new bcmstr:",self.bcmstr
      self.board.WriteBlock("bcmasks.data", self.bcmasks, 0)
    else:
      print "nothing done..."
  def getst(self, ix):
    """
    ix -0..  -points to self.bcmstr
    rc: token,ix   ix -> points just after token
    """
    st= ST(self.bcmstr[ix:])
    return st
  def bcm2bits(self,ix=0):
    """ input: 
    bcm2bits(0) returns:
    ([array_of_bits],string_length_of self.bcmstr)
    """
    self.level= self.level+1
    if self.level>=10:return [2],0
    #print "bcm2bitsSTART:",self.level,ix
    repc=1
    ix1=ix
    bitstr=[]
    while 1:
      st= self.getst(ix1)
      if st.type==ST.BIT:    # push
        for i in range(repc):
          bitstr.append(st.value)
        repc=1
        ix1=ix1+st.length
      elif st.type==ST.REP:
        repc=st.value
        ix1=ix1+st.length
      elif st.type==ST.LEFT:
        ix1=ix1+st.length
        bitstr2,ix2= self.bcm2bits(ix1)
        for i in range(repc):
          bitstr= bitstr+ bitstr2
        repc=1
        ix1=ix1+ix2
      elif st.type==ST.RIGHT:    # pop
        ix1=ix1+st.length
        if self.level<=1:
          PrintError(") incorrectly placed at"+str(ix))       
        break
      elif st.type==ST.EOM:    # pop
        ix1=ix1+st.length
        if self.level>1:
          PrintError(") missing")       
        break
      else:
        PrintError("incorrect BC mask")       
        break
    self.level= self.level-1
    #print "bcm2bits:",bitstr,ix1
    return bitstr,ix1-ix
  def setbits_notused(self,lst=None):
    """lst: [pointer to list, bitnumber]
    bitnumber: 0..3
    """
    bcm,procesed=self.bcm2bits(0)
    if len(bcm)>ORBITLENGTH:
      return "Error: Too long BCmask definition\n"
    if procesed== len(self.bcmstr):
      bitmask= 1<<lst[1]
      #print "setbits1:",bitmask,bcm,procesed
      #lst[0][1:2]=[10,11]
      for ix in range(len(bcm)):
        if bcm[ix]==1:
          lst[0][ix]= lst[0][ix] | bitmask
        else:
          lst[0][ix]= lst[0][ix] & (0xf & ~bitmask)
      #print "setbits2:",lst[0][0:9]
      if len(bcm)<ORBITLENGTH:
        if ErrorMessage=='':
          return "Warning: %d: short BCmask definition for %s (bits above %d not touched)\n"%(
            len(bcm), self.name, len(bcm))
        else:
          return ErrorMessage
      else:
        return None
    else:
      return ErrorMessage
  def checkSyntax(self):
    global ErrorMessage
    ErrorMessage=''
    locbcmasks=[]; 
    for i in range(ORBITLENGTH):
      locbcmasks.append(0)
    errmsg= self.setbits([locbcmasks, 0])
    return errmsg
  def chrep(self, ch, rep):
    if rep==1:
      return ch
    else:
      if len(ch)==1:
        return "%d%s"%(rep, ch)
      else:
        return "%d(%s)"%(rep, ch)
  def shrink(self, msk,width=1,start=0):
    """ correct result only for width=1 with any mask.
    Idea with 'width' seems not applicable:
    H1 -not repeatable, i.e. string must be finished by H, L or )
    in the case of ), the corresponding nmber of ( must match
    examples of good/bad msk:
  
    """
    ix1=start; 
    ous= msk[:start]
    lenmsk=len(msk);
    while ix1<lenmsk:
      ix2= ix1+width; 
      rep=1; 
      if ix2>=lenmsk:
        ch=msk[ix1:]
        ous= ous+self.chrep(ch,rep)
        break
      ch=msk[ix1:ix2]
      #print "bcmask.shrink:%d:ch:"%ix1,ch ; time.sleep(0.1)
      while ix2<lenmsk:
        if ch==msk[ix2:(ix2+width)]:
          rep= rep+1; ix2= ix2+width
          #print "rep:",rep,ix2
        else:
          #ous= ous+self.chrep(ch,rep)
          #ix1=ix2
          break
      ous= ous+self.chrep(ch,rep)
      #print "  ",self.chrep(ch,rep),"ix1:",ix1
      ix1=ix2
    #ous= ous+self.chrep(ch,rep)
    return ous

