#!/usr/bin/python
# 1 ipbus slave:
#  <node id="outcnts" address="0x3600" description="Counters for outgoing signals" fwinfo="endpoint;width=9">
#    <node id="addr" address="0x0"/>
#    <node id="data" mode="port" size="0x200" address="0x1"/>
#  </node>
# seems above not valid, now: outcns.TTcnts: block of 32 counters

import time #,pdb
import hexdump
def prtErr(msg):
  print "ERROR:%s"%msg
def bit_not(n, numbits=32):
  """set 0s in low numbits bits of n"""
  return (1 << numbits) - 1 - n

class Counters:
  """
parent of CountersCTPEMU, CountersTTC, CountersLECO, CountersGBT
                used from
CountersCTPEMU  ltu/ctp
CountersTTC     ltu
CountersLECO    ltu
CountersGBT     ltu
CountersTTCIT   aboard (len:0x40)
"""
  #SIZE= 0    # virtual
  PERLINE= 2 # number of counters in 1 line when displayed
  mymsg= "Output counters."
  def __init__(self, board, ipsl,devname=None, ttccs=None):
    self.ipsl= ipsl
    #self.addrname= ipsl+".addr"
    #self.csa_slave= ipsl+"ctrl"
    if devname==None:
      self.devname= ipsl
    else:
      self.devname= devname
    self.board= board
    self.ttccs= ttccs
    #self.fromad= 0; self.nws= 10; 
    self.last= self.readRaw()   # last readings
    print "%s counters read in (SIZE:%d)."%(self.devname, self.SIZE)
    self.incs= []               # (last - previous) readings
    for ix in range(self.SIZE):
      self.incs.append(0)
  def dodif32(self, before, now):
    if now >= before: 
      dif= now-before
    else: 
      dif= now+ (0xffffffff-before) +1
    #if(DBGcnts) printf("dodif32:%d\n", dif)
    return dif
  def incsupdate(self, block):
    for ix in range(self.SIZE):
      cv= block[ix]
      self.incs[ix]= self.dodif32(self.last[ix], cv)
      self.last[ix]= cv
  def readInc(self):
    prevtime= self.lasttime
    if self.ttccs: ttcblock= self.ttccs.readRaw()
    block= self.readRaw()
    self.inctime= self.lasttime-prevtime
    if self.ttccs: self.ttccs.incsupdate(ttcblock)
    self.incsupdate(block)
    return self.inctime
  #def getlast1(self, ix):
  #  return self.last[ix] 
  def getinc1(self, ix):
    return self.incs[ix] 
  def dumpdata(self, fn, blok):
    os.chdir(os.path.join(os.getenv('HOME'),'WORK'))
    #fp= os.path.join(TRGDBDIR, "ctpinputs_built.cfg")
    f= open(fn, "w")
    for ix in range(len(blok)):
      f.write("%x\n"%blok[ix])
    f.close()
    print "dumpdata: ~/WORK/%s written"%fn
  def prtall(self, names=False):
    print "%s:"%self.devname
    ix= 0; line= ""
    for cn in self.cnames_sorted:
      if names:
        cnta= self.cnames[cn][1]
        lname= self.cnames[cn][0]
        if lname=="":   # (ttcit -all conters are "")
          line= line+ "%25s: %2d"%(cn, cnta)
        else:
          line= line+ "%8s: %2d,%25.25s "%(cn, cnta, lname)
      else:
        cnta= self.cnames[cn][1]
        val= self.getinc1(cnta)
        line= line+ "%8s:%10d    "%(cn, val)
      ix= ix+1
      if ((ix%self.PERLINE)==0) or (ix==len(self.cnames)): 
        print line
        line=""
    if self.ttccs: self.ttccs.prtall(names)
  def excmd(self, cmd, stdprint="stdout"):
    """ stdprint: checked only for r rr r2 cmds
    """
    if len(cmd) < 2:
      self.prtHelp()
      return
    N= self.SIZE
    fromad= 0
    if cmd[1]=='rr' or cmd[1]=='r' or cmd[1]=='r2':
      allcs= True
      if len(cmd)>2: N= int(cmd[2]) ; allcs= False
      if len(cmd)>3: fromad= int(cmd[3]) ; allcs= False
      if cmd[1]=='rr':
        cnts= self.readRaw(N, fromad)
        if stdprint=="stdout":
          hexdump.hd_dump(cnts, 0, len(cnts), fromad, dumptyp='d')
        else:
          return(hexdump.hd_dump(cnts, 0, N-1, fromad, dumptyp='d', stdprint="rc")+" raw values")
      elif cmd[1]=='r' or cmd[1]=='r2':
        if cmd[1]=='r2':
          timedif= self.readInc()
          time.sleep(1)
        timedif= self.readInc()
        #if (self.ipsl=="leco1") or (self.ipsl=="ttc") or (self.ipsl=="gbtX"):
        if self.ipsl!="outcnts":
          tds= "%.2fs"%(timedif)
          if stdprint=="stdout":
            #print "Time delta (system): %.2fs"%(timedif)
            print "Time delta (system): %s"%(tds)
            hexdump.hd_dump(self.incs[fromad:], 0, N-1, fromad, dumptyp='d')
          else:
            return(hexdump.hd_dump(self.incs[fromad:], 0, N-1, fromad, dumptyp='d', stdprint="rc")+" time:"+tds)
        else:
          ms= self.getinc1(0)*3.564E6/40078500
          print "Time delta (system): %.2fs, orbits:%.3fms (+-1 orbit)"%(timedif, ms)
          if allcs:
            self.prtall();
          else:
            hexdump.hd_dump(self.incs[fromad:], 0, N-1, fromad, dumptyp='d')
      else:
        self.prtHelp()
    else:
      if self.ipsl=="leco1":
        if cmd[1]=="clear":
          print "1->ctrl.ctrl.bit2 + 0->ctrl.ctrl.bit2 "
          ctrlctrl= self.board.Read("ctrl.ctrl")
          ctrlctrl= ctrlctrl | 0x4
          self.board.Write("ctrl.ctrl", ctrlctrl)
          ctrlctrl= ctrlctrl & 0xfffffffb
          self.board.Write("ctrl.ctrl", ctrlctrl)
        else:        
          self.prtHelp()
      elif self.ipsl[:8]=="outcnts":
        if cmd[1]=="checktick":
          rc= self.checktick()
          print rc,"  (0:ok, 1:bad tick, 2: cannot be checked (e.g. ONU not in))"
      else:
        self.prtHelp()
  def prtHelp(self):
    print """ %s

r [N] [fromad]    -show difference from the last counters reading
                   N      -number of counters (def: all)
                   fromad -start from fromad counter (def: 0)
r2 [N] [fromad]   -as above, but read+wait 1s before read (i.e.
                   counters accumulated during 1s) 
rr [N] [fromad]   -show raw values
checktick         -check tick (only for outcnts)
"""%self.mymsg
    if self.ipsl!="leco1":
      print """

Relative addresses in array of counters: """
      self.prtall(True)
    else:
      print """
clear             -clear all 64 link error counters
"""
class CountersLECO(Counters):
  """
Link Error COunters (for olt1 only with fi above d_00)
each counter is 32 bits counter
"""
  SIZE= 64
  mymsg= "Link error (upstream) output counters."
  # rel_position_in_outcnts or bit number in TriggerType word
  cnames={   # "ShortName":("LongName", rel_position_in_outcnts)
  "le0":("link error 0",0),
  "le1":("link error 1",1)}
  #cnames_sorted= sorted(self.cnames.items(), key=lambda kv: kv[1][0])
  cnames_sorted= sorted(cnames.keys())
  def __init__(self, board, ipsl,devname=None):
    Counters.__init__(self, board, ipsl,devname)
  def readRaw(self, N=SIZE, fromad=0):
    if fromad>=self.SIZE: fromad= 0
    if (fromad+N)>self.SIZE: N= self.SIZE-fromad
    print "countersLECO.readRaw: N:%d fromad:%d"%(N,fromad)
    block= self.board.ReadBlock(self.ipsl+".data",N, fromad)
    self.lasttime= time.time()
    #1.
    #blocko= block
    # ArgumentError: Python argument types in
    #     ValVector_uint32.__getitem__(ValVector_uint32, slice)
    # did not match C++ signature:
    #     __getitem__(uhal::ValVector<unsigned int>, int)
    # 
    # 2. anyhow, seems the block of all cnts have to be readout 
    # to get proper values
    blocko= []
    for ix in range(len(block)):
      blocko.append(block[ix])
    return blocko
  def prtall(self, names= False):
    fromad= 0 ; N= self.SIZE
    hexdump.hd_dump(self.incs[fromad:], 0, N-1, fromad, dumptyp='d')
class CountersCTPEMU(Counters):
  """
ipbus_ctrs_ported in fi
"""
  SIZE= 34   # 32 counters  from fi 12.0: 2 countrs added: bc40 (tick) and clk240
  mymsg= "CTPEMU output counters."
  def __init__(self, board, ipsl,devname=None, ttccs=None):
    """ ipsl: outcnts (ltu)
ipslN (N: 1..18) counters of ctpemuN
"""
    # rel_position_in_outcnts or bit number in TriggerType word
    self.cnames={   # "ShortName":("LongName", rel_position_in_outcnts)
    "ORB":("Orbit",0),
    "HB":("Heart Beat",1),
    "HBr":("HB Reject", 2),
    "HC":("Health Check",3),
    "TPC_sync":("TPC sync",29),
    "TPC_rst":("TPC rst",30),
    "TOF":("TOF rst",31),
    "PH":("Physics",4),
    "PP":("PrePulse",5),
    "CAL":("Calibration",6),
    "SOT":("SOT", 7),
    "EOT":("EOT", 8),
    "SOC":("SOC", 9),
    "EOC":("SOC",10),
    "TF": ("Time Frame",11),
    "bc40": ("BC (tick)",32),
    "clk240": ("240Mhz clock",33)}
    if board.isfiwa(board.FW_LTU) and board.isfiverhe(0x12):
      self.SIZE= 34
    else:
      self.SIZE= 32
      #print self.cnames
      del(self.cnames["bc40"])
      del(self.cnames["clk240"])
    self.cnames_sorted= sorted(self.cnames.keys())
    #cnames_sorted= sorted(self.cnames.items(), key=lambda kv: kv[1][0])
    #print "selfSIZE:",self.SIZE
    Counters.__init__(self, board, ipsl,devname, ttccs)
  def readRaw(self, N=None, fromad=0):
    if N==None: N= self.SIZE
    if fromad>=self.SIZE: fromad= 0
    if (fromad+N)>self.SIZE: N= self.SIZE-fromad
    #print "readRaw: %d counters from:%d"%(N, fromad)
    block= self.board.ReadBlock(self.ipsl+".TTcnts",N, fromad)
    self.lasttime= time.time()
    #1.
    #blocko= block
    # ArgumentError: Python argument types in
    #     ValVector_uint32.__getitem__(ValVector_uint32, slice)
    # did not match C++ signature:
    #     __getitem__(uhal::ValVector<unsigned int>, int)
    # 
    # 2. anyhow, seems the block of all cnts have to be readout 
    # to get proper values
    blocko= []
    for ix in range(len(block)):
      blocko.append(block[ix])
    return blocko
  def checktick(self):
    if self.SIZE<=32:
      return 2   # cannot be checked
    timedif= self.readInc()
    time.sleep(1)
    timedif= self.readInc()
    ticks= self.getinc1(32)
    c240= self.getinc1(33)
    ticks_vs_240= abs(ticks*6 - c240)
    print "abs(ticks*6:%d - clk240:%d)=%d <---max.5 expected."%(ticks*6, c240, ticks_vs_240)
    if ticks_vs_240 < 6:
      return 0   # ok
    else:
      return 1   # bad tick
     
class CountersTTC(Counters):
  """
TTC counters (+busy timers when ready)
"""
  SIZE= 7   # 7 counters
  mymsg= "TTC output counters."
  ttcclist= ["L0pre","L0","L1","Orbit","PP","L1r","CAL"]
  cnames={   # "ShortName":("LongName", rel_pos)
  # ttc.ShortName_count is the name of the counter reg in .xml
  # rel_pos: not used in fi 8_00
  "L0pre":("CTPEMU output",0),
  "L0":("L0/A bsy", 1),
  "L1":("L1/A",2),
  "Orbit":("Orbit",3),
  "PP":("Prepulse",4),
  "L1r":("L1 reject",5),
  "CAL":("TRD calib",6)}
  cnames_sorted= sorted(cnames.keys())
  def __init__(self, board, ipsl,devname=None):
    Counters.__init__(self, board, ipsl,devname)
  def readRaw(self, N=SIZE, fromad=0):
    if fromad>=self.SIZE: fromad= 0
    if (fromad+N)>self.SIZE: N= self.SIZE-fromad
    #print "readRaw: %d counters from:%d"%(N, fromad)
    blocko= []
    for ix in range(fromad, fromad+N):
      regname= self.ttcclist[ix]+"_count"
      blocko.append(self.board.Read(self.ipsl+"."+regname))
    self.lasttime= time.time()
    return blocko
class CountersGBT(Counters):
  """
GBT counters
"""
  SIZE= 3   # Word_count Error_count(errors) Correct_count(corrections)
  ttcclist= ["Word","Error","Correct"]
  cnames={   # "ShortName":("LongName", rel_pos)
  # ShortName is the name of the counter reg in .xml
  # rel_pos: not used in fi 8_00
  "Word_count":("# of GBT packets received",0),
  "Error_count":("# of errors detected (incorrect data received)", 1),
  "Correct_count":("# of corrections",2)}
  PERLINE= 1
  cnames_sorted= sorted(cnames.keys())
  def __init__(self, board, ipsl,devname=None):
    Counters.__init__(self, board, ipsl,devname)
  def readRaw(self, N=SIZE, fromad=0):
    if fromad>=self.SIZE: fromad= 0
    if (fromad+N)>self.SIZE: N= self.SIZE-fromad
    #print "readRaw: %d counters from:%d"%(N, fromad)
    blocko= []
    for ix in range(fromad, fromad+N):
      regname= self.ttcclist[ix]+"_count"
      blocko.append(self.board.Read(self.ipsl+"."+regname))
    self.lasttime= time.time()
    return blocko
class CountersTTCIT(Counters):
  """
ipbus_ctrs_ported in fi
"""
  SIZE= 64
  mymsg= "TTCIT input counters."
  # rel_position_in_outcnts 
  cnames={   # "ShortName":("LongName", rel_position_in_cnts)
  "data_i.clk40_gated":("",0),
  "data_i.adr_d1..8":("",1),
  "data_i.adr_s1..8":("",9),
  "data_i.adr_e":("",17),
  "data_i.adr_a1..14":("",18),
  "data_i.adr_strobe":("",32),
  "data_i.brc_b":("",33),
  "data_i.brc_e":("",34),
  "data_i.brc_d1..4":("",35),
  "data_i.brc_t1":("",39),
  "data_i.brc_t2":("",40),
  "data_i.brc_strobe":("",41),
  "data_i.l1accept":("",42),
  "status_i.div_nrst":("",43),
  "status_i.err_comm":("",44),
  "status_i.err_dbl":("",45),
  "status_i.err_sng":("",46),
  "status_i.ready":("",47)}
  #cnames_sorted= sorted(self.cnames.items(), key=lambda kv: kv[1][0])
  cnames_sorted= sorted(cnames.keys())
  def __init__(self, board, ipsl,devname=None, ttccs=None):
    """ ipsl: ttcitcnts (ttcit_logic_v4.xml)
"""
    Counters.__init__(self, board, ipsl,devname, ttccs)
  def readRaw(self, N=SIZE, fromad=0):
    if fromad>=self.SIZE: fromad= 0
    if (fromad+N)>self.SIZE: N= self.SIZE-fromad
    #print "readRaw: %d counters from:%d"%(N, fromad)
    block= self.board.ReadBlock(self.ipsl+".TTCitcnts",N, fromad)
    self.lasttime= time.time()
    #1.
    #blocko= block
    # ArgumentError: Python argument types in
    #     ValVector_uint32.__getitem__(ValVector_uint32, slice)
    # did not match C++ signature:
    #     __getitem__(uhal::ValVector<unsigned int>, int)
    # 
    # 2. anyhow, seems the block of all cnts have to be readout 
    # to get proper values
    blocko= []
    for ix in range(len(block)):
      blocko.append(block[ix])
    return blocko
