#!/usr/bin/python
# works with node ctpemu in .xml. Obsoletes: assemu disemu slmdefs emufile
import time
import colors,symbreg
import auxfuns

class CTPemu:
  # NAME_rate:("..._n", TTenabit, outcnts_index)
  outcnts="""
  x -> counter index in signames[][2]
  .    not present in signames

  .  0 Orbit
  .  1 Heart Beat
  x  2 HB Reject
  x  3 Health Check
  x 29 TPC sync
  x 30 TPC rst
  x 31 TOF rst
  x  4 Physics
  x  5 PrePulse
  x  6 Calibration
  x  7 SOT
  x  8 EOT
  x  9 SOC
  x 10 EOC
  x 11 Time Frame

signames[signal]: (Nname,TTenabit,countaddr,PLSRbit) 
Nname: "" -ok exists and is: 
                      ctpemu.signal_n    -the limit on number of trigger signals
                      ctpemu.signal_rate -the rate (BCs between triggers)
       None: does not exists
TTenabit: see above or .xml
countaddr: rel,. address of the counter. None: counter does not exist
PLSRbit: None: cannot be connected to pulser
         N:    ctpemu.ctrl bit enabling pulser for this signal
following for v7:
"""
  #signames= {"HBr_orbit":("",2,2),"HBr_rnd":(None,2,None), 
  signames= {"HBr_orbit":("",2,2,None),
    "HC":("",3,3,None), 
    "PH_bc":("",4,4,8), "PH_rnd":("",4,None,8),
    "PP_bc":("",5,5,9), "CAL_bc":("",6,6,10), 
    "SOT_tf":(None,7,7,None), "EOT_tf":(None,8,8,None),   # just TTena and counter positions
    "SOC_tf":(None,9,9,None), "EOC_tf":(None,10,10,None), # _n/_rate regs do not exist
    "TF_orbit":("",11, 11,None),   # TTena.0x800 bit valid from d02
    "TOF_bc":("",31,31,None), "TPC_SYNC_orbit":("",29,29,11),
    "TPC_RST_bc":("",30,30,12) }
  ctrlbits=("reset","started","SOTEOT","LTU_mode","reset_outcnts","?5","?6","?7",
    "P_PH","P_PP","P_CAL", "P_TPC_SYNC","P_TPC_RST","P_edge")
  def __init__(self, board, ltudevs, emuname):
    #if emuname=="ctpemu":
    #  self.tn= emuname    # top node
    self.tn= emuname    # top node
    self.tnttc= None    # top node for old ttc not used
    self.ctrln= self.tn+".ctrl"
    self.board= board
    self.ltudevs= ltudevs   # ltudevs["tdg"], ltudevs["outcnts"],...
    #self.ttena= 0   # all signals disabled (USRena)
    # rather take actual state, taking into account also _n regs
    # check _n regs (i.e. if >0 set USRena bit):
    if (self.board.Read(self.ctrln) & 2)==0:   # emulation off
      self.checksetTFHBr() # also set TF HBr in TTena
      nbits= 0
      for key in self.signames.keys():
        rname,nname,enabit,pulser= self.getnames(key)
        if enabit==None or nname==None: continue
        if self.signames[key][0]!=None:
          n= self.board.Read(nname)
          if n>0: nbits= nbits | (1<< enabit)
      if (self.ttena & 0xfffff07b) != (nbits & 0xfffff07b):  # do not check TF XOX HBr
        print colors.color("CTPemu: TTena:0x%x does not correspond to choosen triggers:0x%x (N>0)"%\
          (self.ttena&0xfffff07b,nbits&0xfffff07b), "fg_red")
        #self.ttena= self.ttena | nbits
    if self.board.nodExists("ttc", topnode=True):
      self.tnttc= "ttc"   # top node for old ttc
      self.ttcctrl= self.board.Read(self.tnttc+".ctrl")
      self.ttcstatus= symbreg.Register("ttcstatus", PPinhibit=0, PPrequest=1,
        PPselect=2, ORBIT=3, ORBITselect=4, ORBITinhibit=5, dataFIFOwrite=6,
        dataFIFOempty=7, dataFIFOfull=8, dataselect=9, L1FIFOempty=10,
        L1FIFOfull=11, L1FIFOread=12, L1select=13, L1FIFOwrite=14,
        TTCbusy=15, TTC3exbusy=16)
    self.sox= "SOTEOT"   # SOTEOT SOCEOC NOSOXEOX
    self.setsoxbits(self.sox)
  def start(self, msg=""):
    ctrl= self.board.Read(self.ctrln)
    if ctrl & 0x2:
      print "CTP emulator is already started, no action."
      return 
    self.board.Write(self.tn+".rst_N_cnts",0xffffffff);   # reset cnts
    self.board.Write(self.tn+".rst_N_cnts",0);
    self.board.Write(self.tn+".TTena", self.ttena);
    self.board.Write(self.ctrln, ctrl|0x2)   # EMU_run bit: 1
    if msg!="": print msg
    print ".rst_N_cnts: 0xf...-> 0, TTena set to 0x%x, EMU_run:1."%(self.ttena)
  def clearemu(self):
    for signam in self.signames.keys():
      if signam=="TF_orbit": continue
      if signam=="HBr_orbit": continue
      self.set_rn(signam, 0, "0")
  def stop(self,clear=False):
    ctrl= self.board.Read(self.ctrln)
    if (ctrl & 0x2)==0:
      print "CTP emulator was already stopped"
    # clear TTena (do not touch XOX TF_orbit and HBr_orbit):
    ttena= self.board.Read(self.tn+".TTena")
    ttena= ttena & 0xf84   # TF EOC SOC EOT SOT HBr
    self.board.Write(self.tn+".TTena", ttena);
    self.board.Write(self.ctrln, ctrl&0xfd)   # EMU_run bit: 0
    print "TTena set to 0 (but TF_orbit XOX HBr_orbit bits kept), EMU_run:0"
    time.sleep(0.050)   # wait 2 TF after stopped
    if clear:
      self.ttena= ttena
      self.clearemu()
  def getnames(self, rnreg):
    """ rc: rate_regname, N_regname, enabit_position, plsr_bit_position
"""
    nname= self.signames[rnreg][0]
    enablebit= self.signames[rnreg][1]
    if nname==None:
      n= None
      rate= None
    elif nname=="":
      n= self.tn+"."+rnreg+"_n"
      rate= self.tn+"."+rnreg+"_rate"
      if rnreg=="TF_orbit": n= None 
    else:
      # should not happen:
      n= self.tn+"."+nname+"_n"
      rate= self.tn+"."+nname+"_n"
    pulser= self.signames[rnreg][3]
    #print "getnames:",rate,n,enablebit,pulser
    return rate,n,enablebit,pulser
  def prt_rn(self, rnreg):
    rname,nname,enabit,pulserbit= self.getnames(rnreg)
    if rname==None:       # limit on trigger number exist?
      ratestrn="    -"
    else:
      ratestrn= "%5d"%self.board.Read(rname)
    if pulserbit!=None:   # pulser connected?
      ctrl= self.board.Read(self.ctrln) 
      pulser= ctrl & (1<<pulserbit)
    else:
      pulser= 0
    if nname==None:       # limit on trigger number exist?
      strn="-"
    else:
      n= self.board.Read(nname)
      if n==0xffffffff:
        strn= "inf"
      else:
        strn= str(n)
    choosen= ' '
    if (enabit!=None) and (self.ttena & (1<<enabit)):   # choosen signal?
      choosen= '*'
    if pulser==0:
      pulsernote= ""
    else:
      pulsernote= "PLSR"
    line= "%14s%c%s %6s %s"%(rnreg, choosen, ratestrn,strn, pulsernote)
    if enabit==None:
      print line
    else:
      edbits= self.board.Read(self.tn+".TTena")
      if (edbits & (1<<enabit)) or (pulsernote=="PLSR"):
        enadiscol= "fg_green"
      else:
        enadiscol= "fg_red"
      #print "prt_rn:", enadiscol, hex(edbits)
      print colors.color(line, enadiscol) 
  def emureset(self, cbit):
    ctrl= self.board.Read(self.ctrln)
    self.board.Write(self.ctrln,ctrl | cbit)
    self.board.Write(self.ctrln,ctrl & (~cbit))
  def enadis(self, enabit, dis0):
    #TTena= self.board.Read(self.tn+".TTena")
    #if dis0==0:
    #  TTena= TTena & (~(1<<enabit))
    #else:
    #  TTena= TTena | (1<<enabit)
    #self.board.Write(self.tn+".TTena", TTena)
    if dis0==0:
      self.ttena= self.ttena & (~(1<<enabit))
    else:
      self.ttena= self.ttena | (1<<enabit)
  def invertpulser(self, rnreg):
    ctrl= self.board.Read(self.ctrln)
    enabit= self.signames[rnreg][3]
    if enabit==None:
      print(colors.color("%s cannot be connected/disconnected to/from pulser"%\
        rnreg, "fg_red"))
      return
    dis0= ctrl & (1<<enabit)
    if dis0!=0:
      ctrl= ctrl & (~(1<<enabit))
    else:
      ctrl= ctrl | (1<<enabit)
    self.board.Write(self.ctrln, ctrl)
  def set_rn(self, rnreg, rate, nn, immediate=False):
    # rate: int  nn:string
    rname,nname,enabit,pulser= self.getnames(rnreg)
    #print "set_rn:", rnreg,rate,nn,rname,nname,enabit
    if (rate!=None) and (rname!=None) and (rname!=""):
      self.board.Write(rname, int(rate))
    if nname!=None:
      if nn:
        if nn=="inf": intn=0xffffffff
        else:
          intn= int(nn)
        if intn==0:
          self.enadis(enabit, 0)   # disable
        else:
          self.enadis(enabit, 1)   # enable
        self.board.Write(nname, intn)
      if immediate:  
        self.board.Write(self.tn+".TTena", self.ttena)
    self.prt_rn(rnreg)
  def checksetTFHBr(self):
    """ set default values only in case EMU is not active!
    """
    tfname= self.tn+".TF_orbit_rate"
    tfrate= self.board.Read(tfname)
    if tfrate != 256:
      self.board.Write(tfname, 256)
      print colors.color("%s: was %d, rewritten to 256"%(tfname, tfrate), "fg_red")
    #hbrname= self.tn+".HBr_orbit_n"
    #hbrn= self.board.Read(hbrname)
    #if hbrn != 0xffffffff:
    #  self.board.Write(hbrname, 0xffffffff)
    #  print colors.color("%s: was %d, rewritten to 0xffffffff"%(hbrname, hbrn), "fg_red")
    hbrname= self.tn+".HBr_orbit_rate"
    hbrn= self.board.Read(hbrname)
    if hbrn != 0:
      self.board.Write(hbrname, 0)
      print colors.color("%s: was %d, rewritten to 0"%(hbrname, hbrn), "fg_red")
    self.ttena= self.board.Read(self.tn+".TTena")
    if (self.ttena & 0x804) != 0x804:   # TF HBr
      print colors.color("TTena was: 0x%x, setting TF and HBr bits (0x804)..."%self.ttena, "fg_red")
      self.ttena= self.ttena | 0x804
      self.board.Write(self.tn+".TTena", self.ttena)
  def status(self):
    ctrl= self.board.Read(self.ctrln)
    TTena= self.board.Read(self.tn+".TTena")
    sctrl="ctrl:0x%x ("%ctrl
    for ixb in range(32):
      if ctrl & (1<<ixb):
        sctrl= sctrl+ self.ctrlbits[ixb]+" "
    sctrl= sctrl+")"
    print "%s\nTTena:0x%x USRena:0x%x SOX/EOX:%s"%(sctrl, TTena, self.ttena, self.sox)
    for key in self.signames.keys():
      self.prt_rn(key)
    # oldttc part:
    if self.tnttc==None: return
    rw= self.board.Read(self.tnttc+".ctrl") ; self.ttcctrl= rw
    if rw& 0x2:
      detbusy=1
    else:
      detbusy=0
    mode= (rw>>4)&3
    modet= ["0", "oldTTC","TRD","CPV"][mode]
    print "oldttc ctrl:0x%x = mode:0x%x(%s) DETbusy enabled:%d"%(rw, mode, modet, detbusy)
    LLprot= self.board.Read(self.tnttc+".LLprot")
    L1reject= self.board.Read(self.tnttc+".L1reject")
    if L1reject!=0:
      L1rejectp= "(%.2f%%)"%(100./L1reject)
    else:
      L1rejectp= ""
    CALrate= self.board.Read(self.tnttc+".CAL_rate")
    print "       LLprot:%d   L1reject:%d%s   CALrate:%d"%(LLprot, \
      L1reject, L1rejectp, CALrate)
    ttcstatus= self.board.Read(self.tnttc+".status")
    self.ttcstatus.prt(ttcstatus)
  def delme_hbrorbit(self, cmd):
    if len(cmd)==2:
      rate= self.board.Read(self.tn+".HBr_orbit_rate") # period in BCs
      raten= self.board.Read(self.tn+".HBr_orbit_n")   # 0:disbled, 0xffffffff:infinite
      print "%s: %d %d"%(cmd[1], rate, raten)
    elif len(cmd)>=3:
      rate=int(cmd[2])
      self.board.Write(self.tn+".HBr_orbit_rate", rate)
      if len(cmd)==4:
        raten=int(cmd[3])
        self.board.Write(self.tn+".HBr_orbit_n", rate)
  def setttcreg(self,regname, cmd):
    if len(cmd)>3:
      value= eval(cmd[3])
      self.board.Write(regname, value)
      print "%s: 0x%x written"%(regname, value)
      return value
    else:
      #print "%s content missing"%regname
      value= self.board.Read(regname)
      print "%s: 0x%x"%(regname,value)
      return None
  def setsoxbits(self, sox):   # sox: SOTEOT SOCEOC NOSOX
    if sox=="SOTEOT":
      bits_one= (1<<self.signames["SOT_tf"][1]) | (1<<self.signames["EOT_tf"][1])
      bits_zero= ~( (1<<self.signames["SOC_tf"][1]) | (1<<self.signames["EOC_tf"][1]) )
    elif sox=="SOCEOC":
      bits_zero= ~( (1<<self.signames["SOT_tf"][1]) | (1<<self.signames["EOT_tf"][1]) )
      bits_one=  (1<<self.signames["SOC_tf"][1]) | (1<<self.signames["EOC_tf"][1])
    elif sox=='NOSOX':
      bits_zero= ~( (1<<self.signames["SOT_tf"][1]) | (1<<self.signames["EOT_tf"][1]) |\
        (1<<self.signames["SOC_tf"][1]) | (1<<self.signames["EOC_tf"][1]) )
      bits_one= 0
    else:
      print "Bad sox: %s (SOTEOT, SOCEOC or NOSOX expected) "%sox
      return
    ttenabits= self.board.Read(self.tn+".TTena")
    ttenabits= ttenabits & bits_zero
    ttenabits= ttenabits | bits_one
    self.board.Write(self.tn+".TTena", ttenabits);
    self.ttena= ttenabits
    self.sox= sox
  def excmd(self, cmd):
    if len(cmd) < 2:
      self.prtHelp()
    elif cmd[1]=='start':
      self.start()
    elif cmd[1]=='stop':
      clearhw= False
      if len(cmd) > 2:
        if cmd[2]=='clear': clearhw= True
      self.stop(clearhw)
    elif cmd[1]=='reset':
      self.emureset(0x1)
    elif cmd[1]=='resetoutc':
      self.emureset(0x10)
    elif cmd[1]=='status':
      self.status()
    elif cmd[1]=='sss' or cmd[1]=='Tsss' or cmd[1]=='Ssss':
      tdgssm= None
      if cmd[1][0]=='T':
        tdgssm= "tdg"
      if cmd[1][0]=='S':
        tdgssm= "ssm"
      if len(cmd)>=3:
        ms=int(cmd[2])
      else:
        ms= 10
      if tdgssm:
        self.ltudevs[tdgssm].start_recording()
        self.start("recording started on %s..."%(tdgssm))   # start emu
      else:
        self.start()   # start emu
      print("sleep(%d ms)..."%ms)   # start emu
      time.sleep(ms/1000.)
      self.stop()
      if tdgssm:
        time.sleep(0.023)
        print "Stopping recording on %s, after 23ms to get a last TF with EOX"%(tdgssm)
        self.ltudevs[tdgssm].stop_recording()
    elif cmd[1]=='hbr':
      self.prtHelp("hbr")
    elif cmd[1]=='ttc':
      if self.tnttc==None:
        print "ttc part of CTPemu not available."
        return
      mode= 0
      if len(cmd)==2:
        self.prtHelp("oldttc")
        return
      else:
        if cmd[2]=="0": pass
        elif cmd[2]=="status":
          mode= -3
          self.ttcstatus.prtme()
          status= self.board.Read(self.tnttc+".status")
          self.ttcstatus.prt(status)
        elif cmd[2]=="init":
          mode= -3
          self.board.Write(self.tnttc+".chbdata", 0x800006ff) # ttcrx reset
          time.sleep(0.02)
          self.board.Write(self.tnttc+".chbdata", 0x80000000) # fine delay1: 3126ps
          self.board.Write(self.tnttc+".chbdata", 0x80000180) # fine delay2: 16459ps
          self.board.Write(self.tnttc+".chbdata", 0x80000200) # coarse delay: 0
          self.board.Write(self.tnttc+".chbdata", 0x800003f9) # activate B
          self.board.Write(self.tnttc+".chbdata", 0x80018000) # reset BC counter
          print "TTCinit done"
        elif cmd[2]=="oldttc": mode=1
        elif cmd[2]=="trd": mode=2
        elif cmd[2]=="cpv": mode=3
        elif cmd[2]=="busy": mode=-1
        elif cmd[2]=="reset": mode=-2
        elif cmd[2]=="l1reject":
          mode= -3; self.setttcreg(self.tnttc+".L1reject", cmd)
        elif cmd[2]=="llprot":
          mode= -3; self.setttcreg(self.tnttc+".LLprot", cmd)
        elif cmd[2]=="calrate":
          mode= -3; self.setttcreg(self.tnttc+".CAL_rate", cmd)
        elif cmd[2]=="ctrl":
          mode=-3 ; rw= self.setttcreg(self.tnttc+".ctrl", cmd)
          if rw!=None: self.ttcctrl= rw
        else:
          print "bad mode"
          return
      rw= self.board.Read(self.tnttc+".ctrl")
      if mode>=0:
        rw= (rw & 0xff0f) | (mode<<4)
        self.board.Write(self.tnttc+".ctrl", rw) ; self.ttcctrl= rw
      elif mode==-1:
        bsy= (~(rw>>1)&1) & 0x1
        rw= (rw & 0xfffd) | (bsy<<1)
        print "Detbusy now:%d"%bsy
        self.board.Write(self.tnttc+".ctrl", rw) ; self.ttcctrl= rw
      elif mode==-2:
        crst= rw | 0x1
        self.board.Write(self.tnttc+".ctrl", crst)
        rw= rw & 0xfffe
        self.board.Write(self.tnttc+".ctrl", rw) ; self.ttcctrl= rw
        print "reset: write 1 + write 0"
      elif mode==-3:
        pass
    else:
      self.excmdBatch(cmd, help=True)
  def excmdBatch(self,cmd, help=False):
    """ rc: 0 ok
rc>0: error
"""
    rc=0
    if len(cmd) < 2:
      return 0
    if cmd[1]=='trig':
      self.setsoxbits("SOTEOT")
    #elif cmd[1]=='start':
    #  self.start()
    #elif cmd[1]=='stop':
    #  clearhw= False
    #  if len(cmd) > 2:
    #    if cmd[2]=='clear': clearhw= True
    #  self.stop(clearhw)
    elif cmd[1]=='cont':
      self.setsoxbits("SOCEOC")
    elif cmd[1]=='nosox':
      self.setsoxbits("NOSOX")
    elif self.signames.has_key(cmd[1]): 
      signam= cmd[1]
      rate= None; n= "inf"; immediate= False
      if len(cmd)>=3:
        if (cmd[2]=='P') or (cmd[2]=='p'):   # connect to pulser
          self.invertpulser(signam)
          self.prt_rn(signam)
          return
        else:
          rate=auxfuns.fromms(cmd[2])
          if rate==None: return 4   # bad rate argument
      if len(cmd)>=4: n=cmd[3]
      if (len(cmd)>=5) and (cmd[4]=='i'): immediate= True
      self.set_rn(signam, rate, n, immediate)
    else:
      if help: self.prtHelp()
      rc= 8   # bad batch cmd
    return rc
  def prtHelp(self, part=None):
    if part==None:
      print """
emu start                 toggle rst_N_cnts + set TTena bits in hw using USRena + set EMU_run bit
    stop [clear]          0 -> EMU_run + TTena bits in hw
                          clear: clear also hw _rate/_N registers for each trigger signal
    status                CTPemu + oldttc status
    reset                 write 1+0 to bit 0x1 ctpemu.ctrl (reset emu+scalers+HB_counter[0])
    resetoutc             write 1+0 to bit 0x10 ctpemu.ctrl (reset outcnts counters)
    sss [milisecs]        start emulation, sleep(defalut:10 ms), stop emulation
    Tsss [milisecs]         - start TDG recording before sss
    Ssss [milisecs]         - start SSM recording before sss
    trig | cont |nosox    set SOXEOX mode to SOT/EOT, SOC/EOC or noSOX with 
                          start/stop emulation
    SIGNAME [ rate [N [i]] ]
                          -set USRena bit (can be reset by N:0 for a given signal)
                          -set 'SIGNAME rate N' (default N: inf) in hwregs _n,_rate
                          rate:  e.g.for ~orbit rate: 3564 88us 11khz
                             number: distance between 2 signals in BCs (0: 40mhz)
                             number followed by: us(usecs) ms(milisecs) s(secs)   
                                                 hz khz mhz
                          N: number of signals
                            0   -disable,i.e. USRena bit cleared also
                                 (TTena will not be set at next start)
                            inf -infinite number (0xffffffff)
                          i: -> immediate action, i.e. change rate/n during active emulation 
    SIGNAME P             -connect/disconnect SIGNAME to PULSER (ecl). I.e.
                           the resulting rate will be a sum of:
                           - the rate choosen by 'SIGNAME rate N' +
                           - the rate of signals coming over the PULSER input
                           'P' option is allowed only for signals:
                           PH_bc PH_rnd PP_bc CAL_bc TPC_SYNC_orbit TPC_RST_bc
SIGNAME is one of: %s
Note about TOF: TOF can be generated also with a BCmask (OR) see: h bcm

=== Usefull:
outcnts r   -check number of output signals generated from last reading.
             Signals counted BEFORE tdomux!
tdomux      -check/set Trigger Data Output mux (should be emu or emutrd)

emu ttc     -get help for oldTTC
emu hbr     -get help on 'rate' meaning for HBr_orbit signal

"""%(" ".join(self.signames.keys()))
    elif part=='oldttc':
      if self.tnttc==None:
        detbusybit= "-"
      else:
        detbusybit= str((self.ttcctrl&0x2)>>1)
      print """
=== oldTTC detectors:    
    ttc    status         -show ttc.status reg
           MODE           -MODE: one of: 0 oldttc trd cpv
           busy           -flip 'DETbusy enabled' bit (now:%s)
           reset          -counters reset + FIFO clear
           init           -TTCinit (i.e.: ttcrx reset, set fine delays,
                           activate B-channel, resetBC)
           llprot VALUE   -L0-L1 protection in BC (min 8)
           l1reject VALUE -reject rate, e.g. 3: 1/3*100= 33.33%%
           calrate VALUE  -e.g.: 2:each 3rd PH trigger is CAL trigger
           ctrl VALUE     -0xDDAABC  DD:L0 delay in BCs
                           AA: TTC-A delay 01:2/6 02:3/6 04:4/6 08:5/6
                                           10: no delay 20:1/6
                           B:mode 1:oldTTC 2:TRD
                           C:ctrl 1:counter reset+FIFOclear 2:enable DET busy

"""%(detbusybit)
    elif part=='hbr':
      print """HBr_orbit rate: is in interval 0..256. The meaning:
1. TF trigger during an active run (i.e. between SOX/EOX) is never accompanied with HBr bit
2. HBr are generated at the end of TF, i.e.:
   T00....1T00....1   -for rate=255 (T: TF, 0:HBa   1:HBr)
   T00...11T00...11   -for rate=254
3. Number of HBr triggers in a TF: 256-rate, i.e.:

rate  Number of HBr triggers in TF
----  ----------------------------
   0  255    (Rule 1.)
   1  255
   2  254
   3  253
 ...  ...
 254  2
 255  1
 256  0
"""
    else: 
      print "help text not available for ",part

