#!/usr/bin/python
# 2 ipbus slaves:
# ssm
# tdg -trigger data generator
# <node id="data" address="0x8" mode="block" size="0x800" description="data" />
# or 0x100*32= 8192 bits:
#   <node id="pram" address="0x2000" description="1kword peephole RAM" fwinfo="endpoint;width=1">
#    <node id="addr" address="0x0"/>
#    <node id="data" mode="port" size="0x100" address="0x1"/>
# ?
import random
import symbreg, auxfuns, hexdump
import time,os
def prtErr(msg):
  print "ERROR:%s"%msg
def bit_not(n, numbits=32):
    return (1 << numbits) - 1 - n

class DDR3:
  DDRLEN= 1024*1024*1024/4   # 1GB in words
  DDR3_TO=5000       # timeout when reading ddr3 (number of loops)
  #DDR3_BLKL=0x800  # in 32bits words = 65536 bits (ddrmem fifo in fpga).
  DDR3_BLKL=0x100000 # in 32bits words (4MB=1MWord). Alignment: dummy r/w for chunks of 
  DDR3_CHUNK=4       # shortest chunk of 4 32bit words can be read/written
  DDR3_CONF_REG0=0
  # 31..23 readonly: 
  # 31..28: mem_init,            rdi_fifo_empty,      rdi_fifo_has_space, full_flag, 
  #         mem_init:1 ->ddr can be used mem_init:0 ->ddr cannot be used
  # 27..24: ddr3_ext_rd_itf_rdy, ddr3_ext_wr_itf_rdy, rst_logic,          rd_done, 
  #     23: wr_done
  #  2.. 0: writeonly: Errors_reset, Logic_reset, DDR3_reset 
  DDR3_CONF_REG1=1   # Start address (0, 1-> 4th word, 2 -> 8th word), DDR3-read operation 
                     # shifted <<3
  DDR3_CONF_REG2=2   # Number of 128bits readings, DDR3-read operation
  DDR3_CONF_REG3=3   # Start address, DDR3-write operation
                     # shifted <<3
  DDR3_CONF_REG4=4   # Number of (128bits) writings, DDR3-write operation
  #DDR3_BUFF_DATA=5  # .data block size=0x800
  DDR3_mem_init=0x80000000
  DDR3_rdi_fifo_empty=0x40000000
  DDR3_rdi_fifo_has_space=0x20000000
  DDR3_full_flag=0x10000000
  DDR3_rd_itf_rdy=0x8000000
  DDR3_wr_itf_rdy=0x4000000
  DDR3_rst_logic=0x2000000
  DDR3_rd_done=0x01000000
  DDR3_wr_done=0x00800000
  RDOFFSET=0   # is implemented directly in ReadFifoDDR()
  def __init__(self, board, ipsl,devname=None):
    self.ipsl= ipsl
    self.ipslave= ipsl+".ctrl"
    self.dataname= ipsl+".data"
    self.csa_slave= ipsl+"ctrl"
    if devname==None:
      self.devname= ipsl
    else:
      self.devname= devname
    self.board= board
    self.status= symbreg.Register("STATUS", mem_init=31, rdi_fifo_empty=30,
      rdi_fifo_has_space=29, full_flag=28,
      ddr3_ext_rd_itf_rdy=27, ddr3_ext_wr_itf_rdy=26, rst_logic=25, rd_done=24,
      wr_done=23)
    self.fromad= 0; self.nws= 10; 
    self.ddr3_reset(checkec=True)
  def ddr3_status(self, msg=None):
    #status= self.board.Read(self.ipslave, self.DDR3_CONF_REG0)
    #self.status.prt(status, "REG0")
    #status2= self.board.Read(self.ipslave, self.DDR3_CONF_REG1)
    #self.status.prt(status2, "REG1")
    status= self.board.ReadBlock(self.ipslave, 5, self.DDR3_CONF_REG0)
    if msg:
      print "%s REG0:0x%x. %s"%(self.devname, status[0], msg)
    else:
      self.status.prt(status[0], "REG0")
      self.status.prt(status[1], "REG1")
      ad= self.board.Read(self.csa_slave+".ssm_address")
      status= self.board.Read(self.csa_slave+".ssm_status")  # 3:G_ovf 2:R_ovf  1:Gbusy 0:Rbusy
      # bit 3:r_ovfl (i.e. generation)
      # bit 2:w_ovfl (i.e. recording)
      # bit 1:busy   (i.e. generation)
      # bit 0:busy   (i.e. recording)
      control= self.board.Read(self.csa_slave+".ssm_control")
      print "%s.ssm_ address/status/control: 0x%x 0x%x 0x%x"%(self.csa_slave, ad, status, control)
      nodename= self.csa_slave+".ssm_r_address"
      if (self.ipsl=="tdg") and self.board.nodExists(nodename):
        add= self.board.Read(nodename)   # current address sent out
        print "        ssm_r_address:0x%x"%(add)
  def ddr3_wrdone(self):
    return 0
    irep=0; status= 0
    while (status & self.DDR3_wr_done)==0:
      if irep>self.DDR3_TO:
        prtErr("ddr3_wrdone timeout: status: 0x%x"%status)
        return 1
      status= self.board.Read(self.ipslave, self.DDR3_CONF_REG0)
      irep = irep+1
    return 0
  def ddr3_rddone(self):
    # ============
    # return 0
    # ===========
    irep=0; status= 0
    while (status & self.DDR3_rd_done)==0:
      if irep>self.DDR3_TO:
        prtErr("ddr3_rddone timeout: status: 0x%x"%status)
        return 1
      status= self.board.Read(self.ipslave, self.DDR3_CONF_REG0)
      irep = irep+1
    return 0
  #------------------------------------------------------
  def ddr4_start_rd(self):
    irep=0; status=self.DDR3_rdi_fifo_empty; 
    while (status & self.DDR3_rdi_fifo_empty)==self.DDR3_rdi_fifo_empty:  # 0x1 & 0x40000...= 0
      if irep>self.DDR3_TO: 
        prtErr("ddr4 is empty: status: 0x%x"%satus)
        return 1 
      status= self.board.Read(self.ipslave, self.DDR3_CONF_REG0)
      irep = irep+1 
    return 0      
#------------------------------------------------------
  def ddr3_reset(self, bits=7, checkec=True):
    # checkec == True: reset only if REG0 != 0xec000000
    #print "reset %d"%bits
    if checkec:
      # seems double read is needed to read correctly (bug in ipbus <-> ddr4 communication?)
      status= self.board.ReadBlock(self.ipslave, 5, self.DDR3_CONF_REG0)
      status= self.board.ReadBlock(self.ipslave, 5, self.DDR3_CONF_REG0)
    else:
      status= [0,0,0,0,0]
    if status[0] != 0xec000000:
      self.board.Write_nd(self.ipslave, bits, self.DDR3_CONF_REG0)  # in vme not needed after power-up
      self.board.Write_nd(self.ipslave, 0x4, self.DDR3_CONF_REG0)   # Errors_reset
      self.board.dispatch()
      self.board.Write_nd(self.ipslave, 0x0, self.DDR3_CONF_REG0)   # Errors_reset, Logic_reset, DDR3_reset
      self.board.dispatch()
      # seems double read is needed to read correctly (bug in ipbus <-> ddr4 communication?)
      for i in range(10):
        time.sleep(0.001)
        status2= self.board.ReadBlock(self.ipslave, 5, self.DDR3_CONF_REG0)
        if status2[0]!= 0xec000000:
          if i>0: print "%d. reading, status:0x%x"%(i,status2[0])
      #vmew32(0x9000+SSMaddress, 0); // to get cleared SSMstatus[1] bit (i.e. no overflow)
      print"%s REG0:0x%x (expected: 0xec000000), reset done, now REG0: 0x%x"%\
        (self.devname, status[0], status2[0])
    else:
      print"%s REG0:0x%x as expected, i.e. no reset done."%(self.devname, status[0])
  def ddr3_resetbc(self):
    if self.ipsl!="tdg":
      print "resetbc available ONLY for tdg. Nothing done."
      return
    cslave= self.csa_slave+".ssm_control"
    rv= self.board.Read(cslave) ; rv= rv | 0x100
    self.board.Write(cslave, rv)   # 1+0, the bit should be kept 0 all the time
    self.board.Write(cslave, rv & 0xfffffeff)
  def ReadFifoDDR(self,N):
    # max. N: self.DDR3_BLKL
    if N>self.DDR3_BLKL: return None
    block= self.board.ReadFifo(self.dataname,N)
    #blocko= block
    # ArgumentError: Python argument types in
    #     ValVector_uint32.__getitem__(ValVector_uint32, slice)
    # did not match C++ signature:
    #     __getitem__(uhal::ValVector<unsigned int>, int)
    # 
    blocko= []
    for ix in range(len(block)):
      blocko.append(block[ix])
    return blocko
    # following: used for debugging in the past...
    #n= N ; blocko= []
    #maxblk= 0x800 # 248 
    #RDOFFSET= 0   # read +RDOFFSET words and ignore first RDOFFSET words in 1st block 
    #while n>=maxblk:
    #  block= self.board.ReadFifo(self.dataname,maxblk+RDOFFSET)
    #  #print "ReadFifoDDR:%d:"%len(block),block
    #  for ix in range(RDOFFSET,len(block)):
    #    blocko.append(block[ix])
    #  n= n-maxblk
    #  RDOFFSET= 0
    #if n>0:
    #  #block= self.nodlist[self.applyTN(name)].readBlock(n+RDOFFSET)
    #  #self.hw.dispatch()
    #  block= self.board.ReadFifo(self.dataname,n+RDOFFSET)
    #  #print "ReadFifoDDR:%d:"%len(block),block
    #  for ix in range(RDOFFSET,len(block)):
    #    blocko.append(block[ix])
    #return blocko
  def ddr3_readblock(self, cra,bl=1):
    """ Read bl*4 words from address cra*4 (in words)
cra: 0: 0   1: 4   2: 8   i.e. addressing in 128bits blocks
bl:  block length in 128bits blocks (max: 0x200)
"""
    #print "=== readblock:",cra, "length in 128bits blocks:", bl
    self.board.Write_nd(self.ipslave, 0x2, self.DDR3_CONF_REG0)   # Logic reset
    self.board.Write_nd(self.ipslave, 0x0, self.DDR3_CONF_REG0)
    self.board.Write_nd(self.ipslave, cra<<3, self.DDR3_CONF_REG1)
    ##self.board.Write_nd(self.ipslave, cra, self.DDR3_CONF_REG1)
    self.board.Write_nd(self.ipslave, bl, self.DDR3_CONF_REG2)
    #--------
    self.board.dispatch()
    # ----------------------------
    rc= self.ddr4_start_rd()   # data copied from DDR4 to buffer
    if rc!=0: return None
    #satus2= self.board.Read(self.ipslave, self.DDR3_CONF_REG0)
    #prtErr("ddr4: status: 0x%x"%satus2)
    #time.sleep(1)
    # ---------
    #print "=== readblock ReadFifo:", bl*self.DDR3_CHUNK  #+self.RDOFFSET
    blk= self.ReadFifoDDR(bl*self.DDR3_CHUNK) #+4)
    #print "readblock blk:", blk
    #return blk[6:]   #ValVector_uint32.__getitem__(ValVector_uint32, slice)
    return blk
  def ddr3_read(self, ddr3_ad, nws, prt=True):
    """ 
ddr3_ad: address in 32bits words, i.e. 0..(1024*1024*1024/4)-1
nws: number of 32bitswords to read
rc: list of 32bits ints of length nws (nws does not have to be multiple of 4)
"""
    # calculate number of big blocks and length of the last block (must be multiple of 4 words):
    mem_ad= []
    blocks= nws/self.DDR3_BLKL; llb= nws % self.DDR3_BLKL; 
    block0= ddr3_ad/self.DDR3_CHUNK; cra= block0
    for iblks in range(blocks):
      blok= self.ddr3_readblock(cra,self.DDR3_BLKL/self.DDR3_CHUNK)   # 0x200= 512 128bits words = 8192 bytes
      #self.board.Write_nd(self.ipslave, cra, self.DDR3_CONF_REG1)
      #self.board.Write_nd(self.ipslave, 1, self.DDR3_CONF_REG2)
      #self.board.dispatch()
      #rc= self.ddr3_rddone(); if rc!=0: return None
      #print("ddr3_read: now 0x100 words from ddr3_ad:0x%x cra:%d ..."%(ddr3_ad, cra))
      #mem_ad= mem_ad+self.board.ReadBlock(self.dataname, self.DDR3_BLKL, 0)
      if blok==None: return mem_ad
      cra= cra+(self.DDR3_BLKL/self.DDR3_CHUNK)
      mem_ad= mem_ad+blok[self.RDOFFSET:]
    if llb>0:       # only part of the last block copied to output data
      blok= self.ddr3_readblock(cra, (llb-1)/self.DDR3_CHUNK+1)
      #self.board.Write_nd(self.ipslave, cra, self.DDR3_CONF_REG1)
      #self.board.Write_nd(self.ipslave, 1, self.DDR3_CONF_REGe
      #self.board.dispatch()
      #rc= self.ddr3_rddone(); if rc!=0: return None
      #mem_ad= mem_ad+self.board.ReadBlock(self.dataname, self.DDR3_BLKL, 0)[:llb]
      if blok==None: return mem_ad
      if prt and (len(blok)<=0x300):
        print "ddr3_read blok:", blok
        print "ddr3_read llb:%d words, mem_ad:"%llb, mem_ad
      #mem_ad= mem_ad+blok[:llb]   nebavi: 
      #ArgumentError: Python argument types in
      #    ValVector_uint32.__getitem__(ValVector_uint32, slice)
      #did not match C++ signature:
      #    __getitem__(uhal::ValVector<unsigned int>, int)
      #mem_ad= mem_ad+map(int,blok[:llb])   nebavi
      #-----------------------------------------------------
      #for ix in range(self.RDOFFSET, llb+self.RDOFFSET):  # -1?
      #  mem_ad.append(blok[ix])  
      mem_ad= mem_ad+blok[self.RDOFFSET:]
      #-----------------------------------------------------
    return mem_ad
  def ddr3_writeblock(self, cra, blok):
    #print "=== writeblock (1block=128bits):",cra
    self.board.Write_nd(self.ipslave, 0x2, self.DDR3_CONF_REG0)   # Logic reset
    self.board.Write_nd(self.ipslave, 0x0, self.DDR3_CONF_REG0)
    self.board.Write_nd(self.ipslave, cra<<3, self.DDR3_CONF_REG3)
    ##self.board.Write_nd(self.ipslave, cra, self.DDR3_CONF_REG3)
    self.board.Write_nd(self.ipslave, len(blok)/self.DDR3_CHUNK, self.DDR3_CONF_REG4)
    # block way:
    #self.board.WriteBlock_nd(self.dataname, blok, 0)
    # fifo way
    # ---------:
    self.board.WriteFifo(self.dataname, blok)   # implies dispatch
    rc= self.ddr3_wrdone()
    if rc!=0: return None
    return 1
  def ddr3_write(self, ddr3_ad, memblock):
    """
ddr3_ad: must be multiple of 4
memblock: list of 32bits ints, length: multiple of 4 (if not, padded with 0s)
rc: number of written words
"""
    nws= len(memblock)
    wordswritten= 0
    ixfrom= 0
    blocks= nws/self.DDR3_BLKL; llb= nws % self.DDR3_BLKL; 
    block0= ddr3_ad/self.DDR3_CHUNK; cra= block0
    for iblks in range(blocks+1):
      if iblks==blocks:   # align the length of the last one on 4-words boundary
        max4= [] ; ix=0
        while ((llb+ix)%4) != 0:
          max4.append(0); ix= ix+1
        if self.ddr3_writeblock(cra, memblock[ixfrom:(ixfrom+llb)]+max4)==None:
          return wordswritten
        lastlen= llb+ ix
      else:
        if self.ddr3_writeblock(cra, memblock[ixfrom:(ixfrom+self.DDR3_BLKL)])==None:
          return wordswritten
        lastlen= self.DDR3_BLKL
      #cra= block0+iblks*(self.DDR3_BLKL/self.DDR3_CHUNK)
      cra= cra+(lastlen/self.DDR3_CHUNK)
      wordswritten= wordswritten+ lastlen
      ixfrom= ixfrom+self.DDR3_BLKL
    return wordswritten
  def checkword(self, ix, w1, exp):
    if w1!=exp:
      if not self.prtsuppressed:
        print "%d: 0x%x exp: 0x%x"%(ix, w1, exp)
        if self.errs>10:
          prtErr("Too many errors (>10), printing suppressed...")
          self.prtsuppressed= True
      return 1
    else:
      return 0
  def checkblock(self, blok, blok2=None):
    self.errs=0; checked= 0
    if blok2:
      for ix in range(len(blok)):
        self.errs= self.errs + self.checkword(ix, blok[ix], blok2[ix])
        checked= checked+1
    else:    
      random.seed(self.rbseed)
      for ix in range(len(blok)):
        if self.seqran=="s":
          exp= ix
        else:
          exp= random.randint(1,0xffffffff)
        self.errs= self.errs + self.checkword(ix, blok[ix], exp)
        checked= checked+1
    return checked,self.errs
  def checkrecconst(self, blok):
    """ Check block retrieved by recording a 'const' pattern generated by ttcpon.
"""
    self.errs= 0 ; checked= 0
    w4= (4023058174, 3703216246, 1412567125, 1431655765)
    for ix in range(len(blok)/4):
      for ix4 in range(4):
        self.errs= self.errs + self.checkword(4*ix+ix4, blok[4*ix+ix4], w4[ix4])
        checked= checked+1
    return checked,self.errs
  def nextw4count(self, byte):
    if byte == 0x9f:   # 'count' pattern wraps around at 0x9f
      byte1= 0
    else:
      byte1= byte+1
    w4= byte1 + (byte1*256) + (byte1*256*256) + (byte1*256*256*256)
    return w4,byte1
  def checkreccount(self, blok):
    """ Check block retrieved by recording a 'count' pattern generated by ttcpon.
"""
    self.errs= 0 ; checked= 0
    w= blok[0] ; byte= w&0xff
    if (byte == ((w>>8) & byte)) and (byte == ((w>>16) & byte)) and (byte == ((w>>24) & byte)):
      print "count pattern starting with 0x%x"%w
      w4= w
    else:
      print "not count pattern:0x%x"%w
      return 0,0
    for ix in range(len(blok)/4):
      for ix4 in range(4):
        self.errs= self.errs + self.checkword(4*ix+ix4, blok[4*ix+ix4], w4)
        checked= checked+1
      w4,byte= self.nextw4count(byte)
    return checked,self.errs
  def dumpdata(self, fn, blok):
    os.chdir(os.path.join(os.getenv('HOME'),'WORK'))
    f= open(fn, "w")
    for ix in range(len(blok)):
      f.write("%x\n"%blok[ix])
    #f= open(fn, "wb")     binary: todo
    #for ix in range(len(blok)):
    #  f.write("%x\n"%blok[ix])
    f.close()
    print "dumpdata: ~/WORK/%s written"%fn
  def start_recording(self):
    #self.board.Write(self.csa_slave+".ssm_address", 0)  is set by fi 
    cslave= self.csa_slave+".ssm_control"
    rv= self.board.Read(cslave) ; rv= rv | 0x1
    self.board.Write(cslave, rv)   # Start: 0->1, the bit should be kept 0 all the time
    self.board.Write(cslave, rv & 0xfffffffe)
  def stop_recording(self):
    cslave= self.csa_slave+".ssm_control"
    rv= self.board.Read(cslave) ; rv= rv | 0x2
    self.board.Write(cslave, rv)
    self.board.Write(cslave, rv & 0xffffffd)
    addr2= self.board.Read(self.csa_slave+".ssm_address")
    status2= self.board.Read(self.csa_slave+".ssm_status")
    print "Stop, ssm_address is:0x%x ssm_status:0x%x"%(addr2, status2)
  def startstop_recording(self, stopms=0, printstdout=False):
    cslave= self.csa_slave+".ssm_control"
    self.start_recording()
    addr= self.board.Read(self.csa_slave+".ssm_address")
    status= self.board.Read(self.csa_slave+".ssm_status")
    if stopms>0: time.sleep(stopms/1000.)
    rv= self.board.Read(cslave) ; rv= rv | 0x2
    self.board.Write(cslave, rv)   # 0x2 bit: 0-1  (and keep it 0 all the time)
    self.board.Write(cslave, rv & 0xfffffffd)
    addr2= self.board.Read(self.csa_slave+".ssm_address")
    status2= self.board.Read(self.csa_slave+".ssm_status")
    if printstdout:
      print "Just after r_start: ssm_address is:0x%x ssm_status:0x%x"%(addr, status)
      print "Stop after %d ms, ssm_address is:0x%x ssm_status:0x%x"%(stopms,addr2, status2)
    return status2,addr2
  def start_generation(self):
    #self.board.Write(self.csa_slave+".ssm_address", 0)  is set by fi 
    cslave= self.csa_slave+".ssm_control"
    rv= self.board.Read(cslave) ; rv= rv | 0x5
    self.board.Write(cslave, rv)    # Start bit0: 0->1, the bit should be kept 0 all the time
    self.board.Write(cslave, rv & 0xfffffffe)     # Keep mode bit2 ON = generation
  def stop_generation(self):
    cslave= self.csa_slave+".ssm_control"
    rv= self.board.Read(cslave) ; rv= rv | 0x6
    self.board.Write(cslave, rv)     # Stop bit1: 0->1
    self.board.Write(cslave, rv & 0xfffffff9)     # Mode bit2: 0 no generation (recording is default)
  def startstop_generation(self, stopms=0):
    self.start_generation()
    addr= self.board.Read(self.csa_slave+".ssm_r_address")
    status= self.board.Read(self.csa_slave+".ssm_status")
    print "Just after gstart: ssm_r_address is:0x%x ssm_status:0x%x, sleeping %dms..."%(addr, status, stopms)
    if stopms>0: time.sleep(stopms/1000.)
    self.stop_generation()
  def sel_tx(self):
    cslave= self.csa_slave+".ssm_control"
    rv= self.board.Read(cslave) ; rv= rv & 0xffffffef
    self.board.Write(cslave, rv)     # bit4: 0: tx (downstream)
  def sel_rx(self):
    cslave= self.csa_slave+".ssm_control"
    rv= self.board.Read(cslave) ; rv= rv | 0x10
    self.board.Write(cslave, rv)     # bit4: 1: rx (upstream)
  def excmd(self, cmd):
    #print "DDR3:%s cmd:"%self.devname, cmd
    self.rbseed= 253
    if len(cmd) < 2:
      self.prtHelp()
      return
    if cmd[1]=='status':
      self.ddr3_status()
    elif cmd[1]=='reset':
      bits=7
      if len(cmd) > 2: bits= eval(cmd[2])
      if (len(cmd) > 3) and (cmd[3]=='force'):
        self.ddr3_reset(bits&7, False)
      else:
        self.ddr3_reset(bits&7)
    elif cmd[1]=='resetbc':
      self.ddr3_resetbc()
    elif cmd[1]=='rd' or cmd[1]=='rf':
      checkdata= ""; self.seqran="" ; filename= None ; 
      validonly= None
      if len(cmd) > 2: self.fromad= eval(cmd[2])
      if len(cmd) > 3: self.nws= eval(cmd[3])
      if len(cmd) > 4: 
        if cmd[4] == 'f':
          filename= "bindata.rd"
        elif (cmd[4] == 'c') or (cmd[4] == 'count') or (cmd[4] == 'const') or (cmd[4] == 'dpon'):
          checkdata= cmd[4]
          if len(cmd) > 5: 
            validonly= cmd[5]
        else:
          print "ERROR: %s (c f count const dpon allowed)"%cmd[4]
      t1= time.time()
      blok= self.ddr3_read(self.fromad, self.nws, prt=False)
      t2= time.time() ; t= t2-t1; average= self.nws*4./(t*1024*1024)
      print "Read: %d words from %d, %ds %.2f MB/sec "%(self.nws, self.fromad, t, average)
      #print "Block %d read "%fromad, str(map(hex,blok[0:8])),"..."
      #print "Block %d read "%fromad, blok[0:8],"..."
      if checkdata == 'dpon':
        hexdump.hd_dump(blok, 0, len(blok), self.fromad, "dpon", onlyvalid=validonly)
      else:
        limitblocklen= 1000
        if len(blok)<=limitblocklen:
          #print "%d words read from %d: "%(len(blok),self.fromad), map(hex,blok)
          hexdump.hd_dump(blok, 0, len(blok), self.fromad)
        else:
          #print "%d words read from %d "%(len(blok),self.fromad), map(hex,blok[:10]),"...",map(hex,blok[len(blok)-10:])
          hexdump.hd_dump(blok, 0, limitblocklen/10, self.fromad)
          print "..."  #,len(blok)
          hexdump.hd_dump(blok, len(blok)-10, len(blok)-1, self.fromad)
      if checkdata == "c" or checkdata=="count" or checkdata=="const":
        self.prtsuppressed= False
        if checkdata=="count":
          checked,errs= self.checkreccount(blok)
        elif checkdata=="const":
          checked,errs= self.checkrecconst(blok)
        else:
          if blok[0]==0:
            print "Checking sequential data..."
            self.seqran="s"
          else:
            print "Checking random data..."
            self.seqran="r"
          checked,errs= self.checkblock(blok)
        t3= time.time() ; t= t3-t2; average= self.nws*4./(t*1024*1024)
        print "checked: %d words errs:%d %ds %.2f MB/sec "%(checked, errs,t, average)
      if filename: self.dumpdata(filename, blok)
    elif cmd[1]=='rdrep':
      fromad= self.fromad; nws= self.nws; checkdata= "rdrep" ; nrep= 1
      self.prtsuppressed= False
      if len(cmd) > 2: nrep= eval(cmd[2])
      checked=0; errs= 0
      for ixn in range(nrep):
        blok= self.ddr3_read(fromad, nws, False)
        if self.seqran!="":
          checked1,errs1= self.checkblock(blok)
          checked+= checked1; errs+= errs1
      if self.seqran!="":
        print "Repetitions: %d, checked: %d words, errs:%d"%(nrep, checked, errs)
      else:
        print "Repetitions: %d (ssm/tdg rd %d %d)"%(nrep, fromad, nws)
    elif cmd[1]=='wd':
      fromad= 0; words= self.DDR3_BLKL; self.datat= "s"; filename= None
      if len(cmd) > 2: fromad= eval(cmd[2])
      if len(cmd) > 3: words= eval(cmd[3])
      if len(cmd) > 4: 
        if (cmd[4] == 's') or (cmd[4] == 'r'):
          self.datat= cmd[4]
        elif (cmd[4] == 'sf') or (cmd[4] == 'rf'):
          filename= "bindata.wd"
        else:
          print "ERROR: %s (s, r, sf, rf allowed)"%cmd[4]
      blok=[]
      random.seed(self.rbseed)
      if words>16: print "preparing data..."
      for ix in range(words):
        if self.datat == 's':
          exp= ix
        else:
          exp= random.randint(1,0xffffffff)
        blok.append(exp)
      if words>16: print "writing data to DDR4 ..."
      t1= time.time()
      nw= self.ddr3_write(fromad, blok)
      t2= time.time() ; t= t2-t1; average= words*4./(t*1024*1024)
      if len(blok)<=16:
        print "%d words from %d written: "%(nw,fromad), str(map(hex,blok)),"..."
      elif len(blok)<=200:
        print "%d words from %d written, %.2f MB/sec: "%(nw,fromad,average),\
          str(map(hex,blok[0:8])),"...", str(map(hex,blok[len(blok)-8:]))
      else:
        print "%d words from %d written, %ds %.2f MB/sec"%(nw,fromad,t,average)
      if filename: self.dumpdata(filename, blok)
    elif cmd[1]=='wrc':
      blen= self.DDR3_BLKL; N= 1; 
      self.datat= "r"
      #self.DDR3_BLKL; 
      if len(cmd) > 2: N= eval(cmd[2])
      if len(cmd) > 3: blen= (eval(cmd[3]))/4*4   # has to be multiple of 4
      #if len(cmd) > 4: 
      #  if (cmd[4] == 's') or (cmd[4] == 'r'):
      #    self.datat= cmd[4]
      print "preparing a random data blok of %d words..."%blen
      blok=[]; blokinv=[];
      random.seed(self.rbseed)
      for ix in range(blen):
        if self.datat == 's':
          exp= ix
        else:
          exp= random.randint(1,0xffffffff)
        #expinv= ~exp   # 0 -> -1 (not 0xffffffff) !
        expinv= bit_not(exp)
        blok.append(exp); blokinv.append(expinv)
      Nmem= self.DDRLEN/blen
      print "DDR kapacity %d blocks (1block= %d words), last %d words excluded from the test"%\
        (Nmem, blen, self.DDRLEN%blen)
      nfirst= auxfuns.find1rand(Nmem)
      n= nfirst
      checked=0; errs= 0 ; self.prtsuppressed= False
      t1= time.time()
      for Nix in range(N):
        fromad= n*blen
        if Nix%(N/10+1) == 0:
          print "write/read/check %d. (block %d in words:%d)"%(Nix,n,fromad)
        # 1011...
        nw= self.ddr3_write(fromad, blok)
        if nw != blen:
          print "Error: %d written only"%nw
          break
        blokrd= self.ddr3_read(fromad, blen, prt=False)
        checked1,errs1= self.checkblock(blokrd, blok)
        checked+= checked1; errs+= errs1
        # 0100...
        nw= self.ddr3_write(fromad, blokinv)
        if nw != blen:
          print "Error: %d written only"%nw
          break
        blokrd= self.ddr3_read(fromad, blen, prt=False)
        checked1,errs1= self.checkblock(blokrd, blokinv)
        #checked+= checked1; errs+= errs1
        errs+= errs1
        n= (n+nfirst)%Nmem
      t2= time.time() ; t= t2-t1; average= blen*N*4./(t*1024*1024)
      print "Blocks: %d of %d words, %d secs, %.2f MB/sec, checked: %d words, errs:%d"%\
        (N, blen, t, average, checked, errs)
    elif cmd[1]=='fp':
      blen= self.DDR3_BLKL; N= 3; 
      pt= "V" ; val= 0
      if len(cmd) > 2: pt= cmd[2]
      if len(cmd) > 3: val= eval(cmd[3])
      if len(cmd) > 4: blen= (eval(cmd[4]))/4*4   # has to be multiple of 4
      pt= "V"
      print "fp", pt, hex(val), blen, N
      for nix in range(N):
        print "fp block:",nix
        fromad= nix*blen
        blk= self.ddr3_read(fromad, blen, prt=False)
        for nix2 in range(blen/4):
          w0= blk[nix2*4+3]
          #w4= blk[nix2*4]
          if blk[nix2*4] & (1<<23):   # Valid bit
            print "1st valid at: 0x%x"%(fromad+ nix2*4)
            break
    elif cmd[1]=='startstop':
      ms= 0
      if len(cmd) > 2: ms= int(cmd[2])
      # start generation/recording:
      self.startstop_recording(ms, True)
    elif cmd[1]=='gstartstop':
      ms= 0
      if len(cmd) > 2: ms= int(cmd[2])
      # start generation/recording:
      self.startstop_generation(ms)
      addr= self.board.Read(self.csa_slave+".ssm_r_address")
      status= self.board.Read(self.csa_slave+".ssm_status")
      print "Stop generation after %d ms, ssm_r_address is:0x%x ssm_status:0x%x"%(ms,addr, status)
    elif cmd[1]=='gstart':
      self.start_generation()
      status= self.board.Read(self.csa_slave+".ssm_status")
      add= self.board.Read(self.csa_slave+".ssm_r_address")   # current address sent out
      print "Generation started, ssm_status:0x%x ssm_r_address:0x%x"%(status,add)
    elif cmd[1]=='gstop':
      self.stop_generation()
      status= self.board.Read(self.csa_slave+".ssm_status")
      add= self.board.Read(self.csa_slave+".ssm_r_address")
      print "Generation stopped, ssm_status:0x%x ssm_r_address:0x%x"%(status, add)
    elif cmd[1]=='seltx':
      self.sel_tx()
    elif cmd[1]=='selrx':
      self.sel_rx()
    elif cmd[1]=='ttbits':
      self.prtHelpttbits()
    else:
      self.prtHelp()
  def prtHelpttbits(self):
    print """
0 ORBIT ORBIT
1 HB Heart Beat flag
2 HBr Heart Beat reject flag
3 HC Health Check
4 PhT Physics Trigger
5 PP Pre Pulse for calibration
6 Cal Calibration trigger
7 SOT Start of Triggered Data
8 EOT End of Triggered Data
9 SOC Start of Continuous Data
10 EOC End of Continuous Data
11 TF Time Frame delimiter   (0x800)
... Spare
29 TPCsync TPC synchronisation
30 TPCrst TPC reset
31 TOF TOF special trigger
"""
  def prtHelp(self):
    print """
status           -show status register (0xec000000 expected) + tdgctrl. registers
                  tdgctrl.ssm_status bits: 3:G_ovf 2:R_ovf  1:G_busy 0:R_busy
                  Generation/Recording_overflow/busy
                  R_ovf: cleared by fi always when recording started
reset [N [force]]        -ddr4 reset   (default N: 7)
resetbc          -ddr4 resetbc( '1+0' into bit 8 (0x100) of tdgctrl.ssm_control). Has sense only for TDG!
                  N bits: 2..0: Errors_reset, Logic_reset, DDR3_reset
seltx            -record downstream (tx) 0x10 bit: 0 in tdgctrl.ssm_control
selrx            -record  upstream  (rx) 0x10 bit: 1 in tdgctrl.ssm_control

startstop [ms]   -start recording, sleep milsecs, stop recording 
                  time estimation for 1Gbyte of DDR, 128bits/25ns:
                  (2**30/16)*25= 1.677721600 sec
                  See ssm_address how far the recording went.

gstart           -start generation, i.e. data read and sent:
                  A. over TTCPON/GBT (tdg device, tdomux set to tdg), or
                     see ssm_r_address how far the generation proceeded
                  B. replacing data received from ONU (tdimux set to ssm)
                     not implemented yet (as of fi a_0)
                  TDOMUX (tdomux tdg) has to be switched to TDG after tdg generation started!
gstop            -stop generation 
gstartstop [ms]   -start generation, sleep milsecs, stop generation 
                   Only for debugging:be aware tdomux can be switched 
                   to tdg only after gstart

wd addr words [s|sf|r|rf] -write data from DDR4 address addr 
     addr:  in words, i.e. min:0    max:1024*1024*1024/4-1 = 268435455
     words: number of words (32 bits), max: 268435456 = 0x10000000
     s:     sequential data generated  sf: +write to file bindata.wd
     r:     random data generated  rf: + write to file bindata.wd

rd addr words [c | f | count | const | [dpon [pattern]] ]   -read  data from DDR4 address addr
     words: number of words (32 bits)
            1 TF = 3564*256*4 = 3649536 words (readout needs cca 15 secs)
            1/((1/40.079)*3564) = 11246 orbits/s, ie TF= 256/11246 = cca 22.8ms
            1 orbit = 3564*4= 14256 words
     c:     check data. If requested, data should be written before (s or r)
            to the same address.
     f:     do not check, but write to file bindata.rd
     count: check after TTCPON data recorded (i.e. set 'count' pattern in OLT SEL_TX_PATTERN
            and 'startstop before'). Expected pattern:
            ... 0x9e9e9e9e 0x9f9f9f9f 0x0 0x01010101 ...
     const: check TTCPON 'const' pattern, i.e.:
            4023058174, 3703216246, 1412567125, 1431655765,
            4023058174, 3703216246, 1412567125, 1431655765,...
  dpon [v]: "downstream ttcpon" data format. v: is 'v' or 0x'hexa':
              v: show only words with Valid bit ON
              0xhexa : show only records where at least 1 bit of 0xhexa
                       pattern is ON in TriggerType bits
                       (Note: 'tdg ttbits' -show Trigger Type bits)
            printed columns:
            Addresss V H WC ORBIT   BCID TTYPE
                         WordCount
                       Header flag
                     Valid bit
            Address- in words (i.e. 1 BCID = 4 words)

rdrep N      - repeat last 'rd' command N times

wrc N blen  -write/read/check N blocks of blen length, pseudo-random address
             -the same random pattern used for each block
             -each block written/read/checked 2 times (using random and inverted data)
             -time: cca 9secs for one 4MB block
             -N: default 1
                 blen     max.N
                 0x100000 0x100   transfer in 4MB blocks (default)
                 0x800    0x20000    
                 0x100    0x100000  
                 4        0x4000000
             -blen: block length in words, default: 0x100000=4MBs
fp pt pattern blen  -find pattern, reading blocks of 'blen' length (in words, default: 0x100000)
                     pattern: 32 bit number (hexa or decimal)
                     pt: pattern type:   V -valid bit (119) in downstream ttcpon data
                              
"""

