#!/usr/bin/python

import time #,pdb
import hexdump,colors
def prtErr(msg):
  print "ERROR:%s"%msg
def bit_not(n, numbits=32):
    return (1 << numbits) - 1 - n

class Sysmon:
  """
FPGA monitoring. 
ug580-ultrascale-sysmon.pdf, atb: SYSMONE1
ADC: 10 bits, 0.2msps
['temperature', 'vccaux', 'vccint', 'vccbram', 'alarms']
0:   a042   9a07   5134   511b   8100  8000 ? 0100: Vuser0 upper
                                 8100 -always ON
                intermitent:     8182  0002: VCCINT upper+lower
                                 818a  0008:VCCBRAM upper+lower
                                 ..8.  OR of ...7f (7 bitov) ok
                                 .1..  Vuser0 upper+lower
alarms: see p.90 22 
        from 9_02 should read 0
vccint  -feeding fpga core
vccbram -feeding bram memory blocks inside fpga core
   aux  -aux voltage

alarms: not sticky (in fpga ver 9_02) ! -i.e. difficult to catch if intermittent
"""
  def __init__(self, board, devname):
    self.ipsl= "ctrl"
    self.devname= devname
    self.board= board
    self.regnames= ["temperature", "vccaux", "vccint", "vccbram", "alarms"]
    self.regs= [0, 0, 0, 0, 0]
  def readAll(self):
    for ix in range(len(self.regnames)):
      rn= self.ipsl+".fpga_"+self.regnames[ix]
      self.regs[ix]= self.board.Read(rn)
  def tempDec(self, adc16):
    # p.40: 10 MSB, on-chip reference
    return ((adc16>>6)*501.3743)/1024-273.6777
  def powDec(self, adc16):
    # p.41: 16 bits:
    return 3*adc16/65536.
  def gettemp(self):
    rn= self.ipsl+".fpga_"+self.regnames[0]
    self.regs[0]= self.board.Read(rn)
    return self.tempDec(self.regs[0])
  def set10(self, regname, bit):
    reg= self.board.Read(regname)
    reg= reg | (1<<bit)
    self.board.Write(regname, reg)
    reg= reg & (~(1<<bit))
    self.board.Write(regname, reg)
    print "%s bit%d set to 1 and 0"%(regname,bit)
  def prtTempVolts(self):
    self.readAll()
    #hexdump.hd_dump(self.regs, 0, len(self.regs)-1)
    temp= self.tempDec(self.regs[0])
    txc= "fg_green"
    if temp>48: txc= "fg_red"
    line= "fpga Temp:%8.4fC Volts:"%temp
    for ix in range(1,4):
      line= line + "%8.4f"%self.powDec(self.regs[ix])
    print colors.color(line, txc)
  def excmd(self, cmd):
    if len(cmd) < 2:
      self.prtHelp()
      return
    if cmd[1]=='rr':
      reps= 1
      if len(cmd) > 2:
        reps= eval(cmd[2]) 
      print self.regnames
      for rep in range(reps):
        self.readAll()
        hexdump.hd_dump(self.regs, 0, len(self.regs)-1)
        line= "%8.4f"%self.tempDec(self.regs[0])
        for ix in range(1,4):
          line= line + "%8.4f"%self.powDec(self.regs[ix])
        print line
    elif cmd[1]=='reset':
      rn= "ctrl.ctrl" ; bit= 0
      if len(cmd) > 2:
        if cmd[2]=='ipbus':
          bit= 0
        elif cmd[2]=='soft':
          bit= 1
        else:
          pass
      self.set10(rn, bit)
    else:
      self.prtHelp()
  def prtHelp(self):
    print """
rr [rep]             -show %s regs (raw data), repeat rep times
reset ipbus     - ipbus reset ctrl.ctrl[0]
reset soft      - soft ipbus reset ctrl.ctrl[1]
"""%(str(self.regnames))
