#!/usr/bin/python
import sys,time,traceback
import ipbuswrap,saxipbus,i2c,colors,specdevs
#from ipbuswrap import Board 
from auxfuns import isinargs,printError
import spi
import spiatlas as spiatlas

import atb.fpga
import atb.ddrmem
import atb.counters as counters
import atb.ctpemu as ctpemu
import atb.tdomux as tdomux

AllBoards= {}
def printError(incolor, msg=""):
  import colors
  print colors.color(incolor, "fg_red"), msg

class ABoard(ipbuswrap.Board):
  def __init__(self, boardname):
    print "===   ABoard(%s) instance..."%boardname
    ipbuswrap.Board.__init__(self, boardname)
    self.deverr= ""   # or err message
    self.i2cinuse= None ;     self.i2cs= []   # 2x i2c bus:
    # i2fs
    # i2cg
    self.i2cg= None   # i2c bus (list of i2c devices)
    self.tc74= None   # TC74 device
    self.spic= [] ; self.spicinuse= None
    self.atbdevs= {}
    self.checkBase()
  def checkBase(self):   # common (any board) devices
    print colors.color("checkBase: checking IPBUS slaves...", "bold")
    fwinfo="not present"; boardid="not present"
    #if self.board.fwinfo:
    if self.fwinfo:
      #self.fwinfo= self.board.fwinfo
      fwinfo= hex(self.fwinfo)
    #if self.board.nodExists("ctrl.boardid"):
    if self.nodExists("ctrl.boardid"):
      #boardid= self.board.Read("ctrl.boardid") & 0xff
      boardid= self.Read("ctrl.boardid") & 0xff
      boardid= "08-00-30-00-2A-%x"%boardid
    #print "%s: boardid:MAC:%s fwinfo:%s"%(self.board.bname, boardid, fwinfo)
    print "%s: boardid:MAC:%s fwinfo:%s"%(self.bname, boardid, fwinfo)
    tnn= "fpga"   # monitoring
    if self.nodExists("ctrl.fpga_temperature"):
      self.atbdevs[tnn]= atb.fpga.Sysmon(self, tnn)
      self.atbdevs[tnn].prtTempVolts()
      #self.atbdevs[tnn].excmd(("any","rr","1"))
    #self.odevs= []   # 'other' devices (not connected to i2c|spi bus)
    tnn="leds"
    if self.nodExists("ctrl.ledenable"):
      #self.odevs.append(specdevs.Leds(self, "LEDS"))
      self.atbdevs[tnn]= specdevs.Leds(self, "LEDS")
    # i2c:
    if self.nodExists("i2cpll", True) and (fwinfo!="not present"):    # pll should be first
      # Note: in Si5345 class, oltreset() can be used only when also i2cmaintttc (i2cg) done!
      i2cpll= i2c.I2Ccore(self,"i2cpll")
      self.i2cs.append(i2cpll) ; self.i2cinuse= len(self.i2cs)-1
      # add devices
      pll= i2c.Si534x(i2cpll, "si5345")
      i2cpll.addDevice(pll)
      if isinargs("pllin0"): fn= "global"
      elif isinargs("pllin1"): fn= "stdalone"
      elif isinargs("pllnocheck"): fn= "pllnocheck"
      else:
        if self.isfiwa((self.FW_LTU_LOGIC, self.FW_LTU_ITSMFT_LOGIC,self.FW_CTP_LOGIC)+\
          self.FW_TEST_LOGIC):
          fn= "stdalone"   # let's assume
        else:
          if self.isfiwa((self.FW_TTCIT_LOGIC,)):
            fn= "pllnocheck"
          else:
            fn= "pllcheckonly"
        #fn= "ok_ltu_logic.hpy"   #default for CTP lab: ok_ltu_logic.hpy -> pll_in0_global.hpy
        #                                for DET labs: ok_ltu_logic.hpy -> pll_in1_stdalone.hpy
      if fn=="pllnocheck":
        print colors.color("Si5345 config NOT checked (pllnocheck arg, or ttcit_logic).", "fg_red")
      elif fn=="pllcheckonly":
        if pll.check("stdalone", False):
          print colors.color("Si5345 config: IN1/stdalone/OLT", "fg_green")
        elif pll.check("global", False):
          print colors.color("Si5345 config: IN0/global/ONU", "fg_green")
        else:
          print colors.color("Si5345 config: check failed (not IN0 neither IN1 config), use pllin0(=global) or pllin1(=stdalone) parameter!", "fg_red")
          #pll.load("ok_ltu_logic.hpy")
          self.deverr="PLL not initialised"
          return
      else:
        if pll.checkandload(fn):
          print colors.color("Si5345 config: %s, ok"%fn, "fg_green")
    else:
      i2cpll= None
      print colors.color("i2cpll not found.", "fg_red")
    i2cmuxs= (0x74, 0x76)
    if self.nodExists("i2cmainttc", True):  # ttcpon (csaba) i2c
      self.i2cg= i2c.I2Ctpcore(self,"i2cmainttc",i2cmuxs)  # i2c slave on both (ctpltu kcu105)
      #self.Write("i2cmainttc.mux", 11)   # 1..9: olt1..9   10:onu   11:ipbus
      self.i2cg.seti2cttcmux("ipbus")
      print "i2cmainttc bus, i2cttc mux set to ipbus access."
      self.i2cs.append(self.i2cg) ; self.i2cinuse= len(self.i2cs)-1
    elif self.nodExists("i2cmain", True):   # Wishbone i2c
      self.i2cg= i2c.I2Ccore(self,"i2cmain",i2cmuxs)  # i2c slave on both (ctpltu kcu105)
      self.i2cs.append(self.i2cg) ; self.i2cinuse= len(self.i2cs)-1
    else:
      self.i2cg= None
      print "NO i2cmain[ttc] present!"
    if self.i2cinuse:
      if len(self.i2cs)>1:
        # initialize 40mhz tick if needed (i.e. when pll was just loaded above):
        #pll.oltreset todo
        self.i2cinuse= 1   # let leave it on i2cg...
      print "Using %s..."%self.i2cs[self.i2cinuse].i2cslave
    if self.nodExists("ctrl.ctrl"):
      # ctrl.ctrl  bits (27.5.2018):
      # 0: ipbus soft reset
      # 1: nuke   -sort of ipbus reset (weaker)
      ctrl= self.Read("ctrl.ctrl")
      self.Write("ctrl.ctrl", 0)
      ctrl2= self.Read("ctrl.ctrl")
      if (ctrl2 & 0x3) == 0:
        print colors.color("ctrl.ctrl: 0x%x, After 0 written:0x%x -ok"%(ctrl, ctrl2), "fg_green")
      else:
        print colors.color("ctrl.ctrl: 0x%x, After 0 written:0x%x -error (0 expected)"%(ctrl, ctrl2), "fg_red")
    if self.i2cg:
      # add devices on mux_u45 (0x74)
      self.tc74= i2c.TC74(self.i2cg, "tc74", mux=(0x76,0, 0x74,1<<0))   # used by getSlow
      self.i2cg.addDevice(self.tc74)
      self.i2cg.addDevice(i2c.SFP(self.i2cg, "SFP", "ipbus", mux=(0x76,0, 0x74,1<<1)))   # ipbus connection
      # add devices on mux_u59 (0x76)
      # FMC       addr   eeprom type    datasheet
      # ttcpon    0x50   24AA02548T     r3doc/20002124E-254617.pdf
      # FM-S18    0x50   M24C02         r3doc/21202h.pdf
      # dodo             CAT24C02       https://twiki.cern.ch/twiki/pub/ALICE/FMCCards/CAT24C01-D.PDF
      # dodoINV          CAT24C02
      # KAYA      0x50   AT24C01B       r3doc/doc5156.pdf
      # TTC              24LC02B        r3doc/21709c.pdf
      print "fmceeprom not checked..."
      #eeprom= i2c.M24C0x(self.i2cg, "fmceeprom", 0x50, mux=(0x74, 0, 0x76, 1<<0)) #SM-18
      #print "eeprom.getid:",eeprom.getid()
      #if eeprom.getid()!=None:
      #  self.i2cg.addDevice(eeprom)
      #
      # adding onboard SFPs SFP1..14: moved from ltuctl.py 15.11.2019
      chkall= isinargs("all")
      if chkall or ((self.fwinfo>>24) == self.FW_LTU_ITSMFT_LOGIC):
        print "Checking all SFPs (all in parameters or ltu_logic_itsmft)..."
      else:
        #if ((self.fwinfo>>24) == self.FW_LTU_LOGIC) or ((self.fwinfo>>24) == self.FW_CTP_LOGIC):
        print colors.color("ctp/ltu_logic,", "bold")+\
          " checking only 5 SFPs: ONU, OLT1, OLT2, GBT1, GBT2... (use 'all' if needed)"
      for role in self.fwsfps.keys():   # olt1..19 onu gbt1..14
        rolesfp= self.fwsfps[role]      # SFP1..12
        rolemux= self.i2cgmuxes[rolesfp]
        if (self.fwinfo>>24) == self.FW_LTU_ITSMFT_LOGIC:
          sfp= i2c.SFP(self.i2cg, rolesfp, role, mux=rolemux)
          #print "rolesfp:",rolesfp,"role:",role,"rolemux:",rolemux
        else:
          if chkall:
            sfp= i2c.SFP(self.i2cg, rolesfp, role, mux=rolemux)
          else:
            if (role=="onu") or (role=="olt1") or (role=="olt2") or (role=="gbt1") or (role=="gbt2"):
              sfp= i2c.SFP(self.i2cg, rolesfp, role, mux=rolemux)
            else:
              continue
        #print "rolesfp:",colors.color(rolesfp+" "+str(sfp.purpose),"bold"),"role:",role,"rolemux:",rolemux
        if sfp.purpose!=None:
          print colors.color(rolesfp+" ","bold"),"role:",role,"rolemux:",rolemux
        if sfp.purpose:
          self.i2cg.addDevice(sfp)
          #print "ok"
    # spi:
    # check if spi0 available before instantiating spic[xx]!
    # i.e. see: ublaze/ttcpon.py DriverIpbus getNodeNames...
    # ctrl.ctrl bit2: valid for SPI0 (StartUpE3) NOT used from ltu_logic_v2
    # 0:used over SPI  1: not used over SPI
    # spi0/1 on both ltukcu + proto_*:
    if self.nodExists("spi0", True):
      print "spi0 found, creating..."
      self.spic.append(spi.SPIcore(self,"spi0"))  # StartUpE3 spi
      self.spic[0].addDevice(spi.N25Q(self.spic[0], "fm_cfg", "n25q128"))
      self.spicinuse= 0
    if self.nodExists("FLASH_SPI_Ram_0",True):
      print "Flash_SPI_Ram_0 found, creating..."
      self.spic.append(spiatlas.SPIatlas())    # here seems error: also node name should be passed to SPIatlas...
      # assuming that either spi0 or FLASH_SPI exlusively is present
      self.spic[len(self.spic)-1].addDevice(spiatlas.N25Q(self.spic[len(self.spic)-1],
        "fm_cfg", "n25q128",self.getHW(),"FLASH_SPI_Ram_0"))
      # check if ok:
      devid= self.spic[len(self.spic)-1].connected[0].getid()
      print "spi dev id:", devid
      if devid!= None:
        self.spicinuse= len(self.spic)-1
      else:
        del self.spic[len(self.spic)-1]
        print colors.color("Flash_SPI_Ram_0 removed","fg_red")
    if self.nodExists("spi1", True):
      self.spic.append(spi.SPIcore(self,"spi1"))  # spi1 (no StartUpE3)
      self.spic[len(self.spic)-1].addDevice(spi.N25Q(self.spic[len(self.spic)-1],
        "fm_semip", "n25q128"))
      self.spicinuse= len(self.spic)-1
    if self.nodExists("FLASH_SPI_Ram_1",True):
      print "Flash_SPI_Ram_1 found, creating..."
      self.spic.append(spiatlas.SPIatlas())
      # assuming that either spi1 or FLASH_SPI exlusively is present
      self.spic[len(self.spic)-1].addDevice(spiatlas.N25Q(self.spic[len(self.spic)-1],
        "fm_semip", "n25q128",self.getHW(),"FLASH_SPI_Ram_1"))
      # check if ok:
      devid= self.spic[len(self.spic)-1].connected[0].getid()
      print "spi dev id:", devid
      if devid!= None:
        self.spicinuse= len(self.spic)-1
      else:
        del self.spic[len(self.spic)-1]
        print colors.color("Flash_SPI_Ram_1 removed","fg_red")
    if self.nodExists("spiadc", True):
      if self.nodExists("ctrl.adc_test_data"):
        # adc: ONLY if spiadc + ctrl.adc_test_data available in .xml: 
        self.spic.append(spi.SPIcore(self,"spiadc", sslogic=True))
        self.spicinuse= len(self.spic)-1
        self.spic[len(self.spic)-1].addDevice(spi.MAX11661(self.spic[len(self.spic)-1],
          "adc", "MAX11661"))
        print "spiadc added. len(spic):",len(self.spic)
    if self.spicinuse==0:
      print "Using spi%d (StartUpE3)"%self.spicinuse
    elif self.spicinuse>0:
      print "Using spi%d"%self.spicinuse
    else:
      print "No spi ?"
    print "DDR4s:"
    for tnn in ("ssm","tdg"):
      if self.nodExists(tnn, topnode=True):
        self.atbdevs[tnn]= atb.ddrmem.DDR3(self, tnn)
  def geti2cinuse(self):
    if self.i2cinuse!=None:
      return self.i2cs[self.i2cinuse].i2cslave
    else:
      return "None"
  def add_ctpemu(self, emuname):
    if self.nodExists(emuname, topnode=True):
      if emuname=="ctpemu":
        emun=""
      else:
        emun=emuname[6]
      try:
        ctpemulator= ctpemu.CTPemu(self,self.atbdevs, emuname)
        self.atbdevs["emu"+emun]= ctpemulator
        print "CTP emulator:", emuname
      except:
        #print sys.exc_info()
        traceback.print_exc(file=sys.stdout)
        ctpemulator= None
        #self.deverr= "CTP emulator not available"
        printError("CTP emulator %s not available"%emuname)
    else:
      self.deverr= "CTP emulator %s not available"%emuname
      printError("CTP emulator %s not available"%emuname)
      #raise Exception("CTP emulator not available")
  def add_outcnts(self, tnn, ttccnts=None, cnttype=None):
    if self.nodExists(tnn, topnode=True):
      if cnttype==None:
        self.atbdevs[tnn]= counters.CountersCTPEMU(self, tnn, ttccs=ttccnts) 
      else:
        self.atbdevs[tnn]= cnttype(self, tnn)
      #print "CTPEMU counters:", tnn
    else:
      printError("%s not available"%tnn)
  def printDevices(self):
    """ print: all devices spi* i2c*
"""
    print colors.color("Detected devices:","bold")
    for ix in range(len(self.spic)):
      #if self.spic[ix].spislave != "spiadc":
      #  print "aboard.py printDevices spi skipping(causing error?):", self.spic[ix].connected[0].devname,\
      #    "slave:", self.spic[ix].spislave
      #  continue
      self.spic[ix].printDevs()
    for ix in range(len(self.i2cs)):
      self.i2cs[ix].printDevs()
    print "Other devices:"
    #for ix in range(len(self.odevs)):
    #  print "  "+self.odevs[ix].getid()
    for devname in self.atbdevs.keys():
      print "  "+devname   #self.atbdevs[devname].getid()
  def findDevice(self, devname):
    """ find all devices connected to spi* i2c* or odevs
when last character of devname is '*': rturn the list of all devices
"""
    for ix in range(len(self.spic)):
      if len(self.spic[ix].connected)==0: continue
      if self.spic[ix].connected[0].devname == devname:   # only 1 device on spi!
        return self.spic[ix].connected[0]
    for ix in range(len(self.i2cs)):
      i2cb= self.i2cs[ix]
      for ixc in range(len(i2cb.connected)):
        if i2cb.connected[ixc].devname == devname:
          return i2cb.connected[ixc]
    #for ix in range(len(self.odevs)):
    #  if self.odevs[ix].devname == devname:
    #    return self.odevs[ix]
    for devn in self.atbdevs.keys():
      #print "findDevice:", devn
      if devn == devname:
        return self.atbdevs[devn]
    return None
  def findDeviceSFPS(self):
    sfps= []
    for ix in range(len(self.i2cs)):
      i2cb= self.i2cs[ix]
      for ixc in range(len(i2cb.connected)):
        dn= i2cb.connected[ixc].devname
        if (len(dn) >=4) and ( (dn[:4]=="FSFP") or (dn[:3]=="SFP") ) :
          sfps.append(i2cb.connected[ixc])
    return sfps
  def getSlow(self):
    """ rc: (errmsg,(tstamp,val1,val2,...))
        errmsg: "" -ok
                error message during measurement
"""
    tstamp= time.time()
    errmsg= ""
    try:
      tc74= self.tc74.gettemp()
    except:
      # seems gets stucked here until tc74 not read in 'atb tc74 id'
      tc74= None
      print "atbwrap.py getSlow: empty tc74"
      errmsg= "empty tc74"
    try:
      ft= self.atbdevs["fpga"].gettemp()
    except:
      ft= None
      print "atbwrap.py getSlow: empty fpga temp"
      if errmsg: errmsg=errmsg+"\n"
      errmsg= errmsg+"empty tc74"
    return (errmsg,(tstamp, tc74, ft))

class CTPBoard(ABoard):
  def __init__(self, boardname):
    print "=== CTPBoard(%s) instance..."%boardname
    ABoard.__init__(self, boardname)
    if self.deverr !="": return
    print "=== CTPBoard(%s) instance checkDevices..."%boardname
    self.checkDevices()   # LTU devices
  def checkDevices(self):
    maxemu=3
    for ix in range(1,maxemu+1):
      self.add_ctpemu("ctpemu%d"%ix)
    for ix in range(1,maxemu+1):
      self.add_outcnts("outcnts%d"%ix)
    tnn= "tdomux"
    self.atbdevs[tnn]= tdomux.CTPTDOmux(self)
    print "TDO mux:", tnn
class TTCITBoard(ABoard):
  def __init__(self, boardname):
    print "=== TTCITBoard(%s) instance..."%boardname
    ABoard.__init__(self, boardname)
    if self.deverr !="": return
    print "=== TTCITBoard(%s) instance checkDevices..."%boardname
    self.checkDevices()   # TTCIT devices
  def checkDevices(self):
    self.add_outcnts("ttcitcnts", None, counters.CountersTTCIT)   # ttcit
class LTUBoard(ABoard):
  def __init__(self, boardname):
    """ todo:
stdalone: tdimux: det1..det3, tdomux: CTPemu
global:   tdimux:n/a          tdomux: onu data
"""
    print "=== LTUBoard(%s) instance..."%boardname
    #ipbuswrap.Board.__init__(self, boardname)
    ABoard.__init__(self, boardname)
    if self.deverr !="": return
    print "=== LTUBoard(%s) instance checkDevices..."%boardname
    self.checkDevices()   # LTU devices
    #if self.isfiwa(self.FW_LTU_LOGIC):
    #  self.checkLtuDevices()
    #elif self.isfiwa(self.FW_CTP_LOGIC):
    #  self.checkCtpDevices()
    #else:
    #  printError("Only ctp_ or ltu_logic_ allowed")
  def checkDevices(self):
    if self.nodExists("i2cfmcsfp", True):
      i2fs= i2c.I2Ccore(self,"i2cfmcsfp")  # only ctpltu
      self.i2cs.append(i2fs)  # ; self.i2cinuse= len(self.i2cs)-1
    else:
      i2fs= None
    # GBT SFPs on FMC connected over i2cfmcsfp:
    if i2fs:          # Front view channel numbers (to be tested)
      if (self.fwinfo>>24) == self.FW_LTU_ITSMFT_LOGIC:
        ##i2fs.addDevice(i2c.SFP(i2fs, "FSFP0", mux=(0,0)))   # 3/228  7/227
        ##i2fs.addDevice(i2c.SFP(i2fs, "FSFP1", mux=(0,1)))   # 2/228  6/227
        ##i2fs.addDevice(i2c.SFP(i2fs, "FSFP2", mux=(0,2)))   # 1/228  5/227
        ###i2fs.addDevice(i2c.SFP(i2fs, "FSFP3", mux=(0,3)))  # 0/228  4/227/not connected
        ##i2fs.addDevice(i2c.SFP(i2fs, "FSFP4", mux=(0,4)))   # not connected ?
        # 13 gbts:
        #i2fs.addDevice(i2c.SFP(i2fs, "FSFP5", "gbt9",mux=(0,5)))
        #i2fs.addDevice(i2c.SFP(i2fs, "FSFP6", "gbt10",mux=(0,6)))
        #i2fs.addDevice(i2c.SFP(i2fs, "FSFP7", "gbt11",mux=(0,7)))
        # 14 gbts:
        sfp= i2c.SFP(i2fs, "FSFP0", "gbt9",  mux=(0,0))
        if sfp.purpose: i2fs.addDevice(sfp)
        sfp= i2c.SFP(i2fs, "FSFP1", "gbt10",  mux=(0,1))
        if sfp.purpose: i2fs.addDevice(sfp)
        sfp= i2c.SFP(i2fs, "FSFP2", "gbt11",  mux=(0,2))
        if sfp.purpose: i2fs.addDevice(sfp)
        sfp= i2c.SFP(i2fs, "FSFP3", "gbt12",  mux=(0,3))
        if sfp.purpose: i2fs.addDevice(sfp)
      else:
        print "FMC devices not checked (not itsmft logic)"
    tnn= "ttccnts" ; ttccnts= None
    if self.nodExists("ttc", topnode=True):
      try:
        self.atbdevs[tnn]= ttccnts= counters.CountersTTC(self, "ttc", tnn)
        print "TTC counters:", tnn
      except:
        if self.atbdevs.has_key(tnn): del self.atbdevs[tnn]
        traceback.print_exc(file=sys.stdout)
    else:
      printError("%s not available"%tnn, "(seems old version of .xml or firmware)")
    self.add_outcnts("outcnts", ttccnts)
    #
    tnn= "leco1"
    if self.nodExists(tnn, topnode=True):
      try:
        self.atbdevs[tnn]= counters.CountersLECO(self, tnn) 
        print "Link error (upstream) counters:", tnn
      except:
        if self.atbdevs.has_key(tnn): del self.atbdevs[tnn]
        traceback.print_exc(file=sys.stdout)
    else:
      printError("%s not available"%tnn, "(seems old version of .xml or firmware)")
    #
    tnn= "tdomux"
    self.atbdevs[tnn]= tdomux.TDOmux(self)
    print "TDO mux:", tnn
    tnn= "tdimux"
    self.atbdevs[tnn]= tdomux.TDImux(self)
    print "TDI mux:", tnn
    self.add_ctpemu("ctpemu")
def Init():
  return ipbuswrap.Init()
def OpenATBoard(bname, role=None):
  """ role: class name: XXXBoard (CTP, LTU or AAAAA)
            None: use alice_cf file and .xml name to to choose the instance
      rc: None if board not found
"""
  if AllBoards.has_key(bname):
    #print "=== OpenATBoard(%s) using existing instance."%bname
    pass
  else:
    if role!= None:
      binstance= role(bname)
    else:
      atable= saxipbus.findAddrTable(bname)
      if atable==None:
        return None
      if atable.find("ltu_logic") == 0:
        binstance= LTUBoard(bname)
      elif atable.find("ctp_logic") == 0:
        binstance= CTPBoard(bname)
      elif atable.find("ttcit_logic") == 0:
        binstance= TTCITBoard(bname)
      else:
        binstance= ABoard(bname)
    AllBoards[bname]= binstance
  return AllBoards[bname]

def main():
  if len(sys.argv) < 2 or (len(sys.argv)==2 and sys.argv[1]=="help"):
    print """
Usage:
atbwrap.py atb_name

"""
    sys.exit()
  phbname= sys.argv[1]
  ipbuswrap.Init()
  atb= OpenATBoard(phbname)
  print "tc74 fpga temperatures",atb.bname,atb.getSlow()

if __name__ == "__main__":
  main()

