#!/usr/bin/python
"""
5.6.2017: GenDev.setmux: virtmux added
8.6.: setmux() invoked in each device method!
"""
import types,string,time,os,math,colors
import hexdump
class ponlogger:
  def info(self, msg):
    print msg
  def warn(self, msg):
    print "WARN:", msg
class GenI2C:
  def __init__(self, board, i2cslave, muxes=None):
    self.board= board
    self.i2cslave= i2cslave   # must be connected
    self.muxes= muxes   # list of addresses of MUXes on this bus
    self.connected= []   # connected devices
  def addDevice(self, device):
    #self.connected.append(device) ; return
    try:
      pn= device.getid()
    except:
      pn= None
    if pn==None:
      print "Not connected/faulty:", device.devname
      return None
    else:
      self.connected.append(device)
      #print "GenI2C: %s ok"%device.devname
      return len(self.connected)-1
  def findDevice(self, devname, index=None):
    if index!=None: return self.connected[index]
    for ix in range(len(self.connected)):
      dev= self.connected[ix]
      if dev.devname== devname: return dev
    return None
  def printDevs(self):
    devnames=""
    print "%s:"%(self.i2cslave)
    ndevs= len(self.connected)
    for ix in range(ndevs):
      devid= self.connected[ix].getid()
      #namid= "%s:%s"%(self.connected[ix].devname,self.connected[ix].getid())
      namid= "%s:%s"%(self.connected[ix].devname, devid)
      #devnames= devnames+"%s:%s"%(self.connected[ix].devname,self.connected[ix].getid())+" "
      print "  %s"%(namid)
    if ndevs==1:
      print "  %d device connected"%(ndevs)
    elif ndevs>1:
      print "  %d devices connected"%(ndevs)
    else:
      print "  No device detected"
  def getMuxes(self):
    if self.muxes: 
      mxs= ""
      for ix in range(0,len(self.muxes)):
        ma= self.muxes[ix]
        if ma==0:   # mux implemented in ctpltu fpga
          mv= self.board.Read("ctrl.fmcsfpsel")
        else:
          mv= self.i2cr(ma)   # get mux
        mxs= mxs+ " 0x%x:0x%x"%(ma, mv)
      return mxs
  def getSTATUS(self,stat=None):
    mxs= self.getMuxes()
    if stat:
      return "%s STATUS: MUXs:%s"%(self.i2cslave, stat, mxs)
    else:
      return "%s STATUS: MUXs:%s"%(self.i2cslave, mxs)
  def prtSTATUS(self, stat=None):
    return self.getSTATUS(stat=None)
class I2Ctpcore(GenI2C):
  """ Usage:   (i2c from ttcpon team)
  # i2c ops:
  b.i2ci()
  b.i2cw(addr, val)  - val: 8bits
  b.i2cwm(addr, memloc, values):      max. blen: 2
  values= b.i2cr(addr, blen=1):
  values= b.i2crm(addr,memloc, blen=1):
"""
  WR= 0x80000000
  RD1= 0xc0000000
  RD2= 0xe0000000
  def __init__(self, board, i2cslave="i2cmainttc", mxs=None):
    GenI2C.__init__(self, board, i2cslave, mxs)
    self.reg= "%s.reg"%(self.i2cslave)
    self.i2cmaindbg= False
  def i2ci(self):   # 100khz, no need to init
    pass
  def checkerr(self, reg):  # used only after 1st write 
    rc=self.board.Read(self.reg)
    if (rc>>16) == 0xaaaa: return rc   # err: rd1/2 wr
    #if (rc>>16) != 0x5555:
    if rc == 0xb0b0b0b0:   # too early, repeat write
      for ixn in range(4):
        time.sleep(0.003)
        #self.board.Write(self.reg, reg)
        rc=self.board.Read(self.reg)
        if rc == 0xb0b0b0b0:  continue
        if (rc>>16) == 0x5555: return rc& 0xffff
        break
      return rc
    else:
      if (reg>>28) == 0x8:
        rc= 0
      elif (reg>>28) == 0xc:
        rc= rc & 0xff
      elif (reg>>28) == 0xe:
        rc= rc & 0xffff
      else:
        rc=0xffff0000
    # rc ok: 0x0000bbaa : bbaa - read value (0 in case of i2cwr)
    # rc err: 0xxxxxbbaa  xxxx: != 0
    return rc
  def seti2cttcmux(self, usedby="ipbus", itsmft=False):
    """ ltu_logic_itsfmt: only olt onu or ipbus allowed
    """
    if usedby[:3]=="olt":
      muxv= int(usedby[3])
      maxv=9
      if itsmft: maxv=1
      if not ((muxv>=1) and (muxv<=maxv)):
        print "Warning: seti2cttcmux: %s -no action (allowed: olt1..9 onu ipbus) or \n olt1 onu ipbus for itsfmt fi "%(usedby)
        return False
    elif usedby[:3]=="onu":
      muxv= 10
    elif usedby=="ipbus":
      muxv= 11
    elif usedby=="h":
      if itsfmt:
        print "Allowed: olt1 onu ipbus"
      else:
        print "Allowed: olt1..9 onu ipbus"
      return False
    else:
      print "Warning: seti2cttcmux: %s -no action (allowed: olt1..9 onu ipbus) or \n olt1 onu ipbus for itsfmt fi "%(usedby)
      return False
    #print "seti2cttcmux: %s %d"%(usedby, muxv)
    #self.b.board.Write("i2cmainttc.mux", muxv)
    self.board.Write("i2cmainttc.mux", muxv)
    return True
  def i2cwr(self, addr, memloc, value):
    reg= I2Ctpcore.WR | (addr<<16) | (memloc<<8) | (value&0xff)
    self.board.Write(self.reg, reg)
    #rc=self.board.Read(self.reg)
    #if (rc>>16) != 0x5555:
    #  print "i2cwr(%x %x %x) error rc:%x reg:%x"%(addr, memloc, value, rc, reg)
    rc= self.checkerr(reg)
    if rc != 0:
      print "i2cwr(%x %x %x) error rc:%x reg:%x"%(addr, memloc, value, rc, reg)
      raise Exception('i2cwr({},{},{})  reg:{} rc:{}'.format(addr, memloc, value,hex(reg), hex(rc)))
    else:
      if self.i2cmaindbg:
        print "i2cwr(%x %x %x) ok rc:%x reg:%x"%(addr, memloc, value, rc, reg)
  def i2crd1(self, addr, memloc):
    reg= I2Ctpcore.RD1 | (addr<<16) | (memloc<<8)
    self.board.Write(self.reg, reg)
    #rc=self.board.Read(self.reg)
    #if (rc>>16) == 0xaaaa:   # error: aaaa.... if not, ok
    rc= self.checkerr(reg)
    if (rc>>16) != 0:
      print "i2crd1(%x %x) error:%x reg:%x"%(addr, memloc, rc, reg)
      raise Exception('i2crd1({},{})  reg:0x{} rc:0x{}'.format(addr, memloc, hex(reg), hex(rc)))
      return None
    else:
      if self.i2cmaindbg:
        print "i2crd1(%x %x): %x reg:%x"%(addr, memloc, rc, reg)
      return rc&0xff   # 1 byte
  def i2crd2(self, addr, memloc, elog=True):
    reg= I2Ctpcore.RD2 | (addr<<16) | (memloc<<8)
    self.board.Write(self.reg, reg)
    #rc=self.board.Read(self.reg)
    #if (rc>>16) == 0xaaaa:
    rc= self.checkerr(reg)
    if (rc>>16) != 0:
      if elog:
        print "i2crd2(%x %x) error:%x reg:%x"%(addr, memloc, rc, reg)
      #raise Exception('i2crd2({},{})  reg:{} rc:{}'.format(addr, memloc, reg, rc))
      return None
    else:
      if self.i2cmaindbg:
        print "i2crd2(%x %x): %x reg:%x"%(addr, memloc, rc, reg)
      return rc&0xffff   # 2 bytes
  # from outside (compatible with I2Ccore):
  def i2cw(self, addr, value):
    """ value: single value (8bits)
    """
    if type(value) == types.ListType:
      print "Error: i2cw: max 1 byte with i2cmainttc"%len(value)
    else:
      self.i2cwr(addr,0,value)
  def i2cwm(self, addr, memloc, values):
    """ addr:  i2c slave addr
    memloc:    memory location
    values:    single value (8bits) or
               list of values, e.g.: [3,67,244] -> 3 data bytes out
    return:    number of bytes acknoledged
    """
    if type(values) == types.ListType:
      print "Error: i2cwm: max 1 byte with i2cmainttc (blen:%d)"%len(values)
    else:
      self.i2cwr(addr,memloc,values)
  def i2cr(self, addr, blen=1):
    """ single reg. read (multiple readings if blen>1)
    return: list of values or 
            single value in case blen==1
            None in case of error 
    """
    if blen==1:
      val= self.i2crd1(addr,0)
      return val
    else:
      print "Error:i2cr: i2cmainttc: mem read used instead of single read (req:%d bytes)"%blen
  def i2crm(self,addr,memloc, blen=1, si534x=False, elog=True, b2=True):
    """ addr:  i2c slave addr
    memloc:    memory location from where to read
    blen:      number of required readings
    """
    if blen==1:
      val= (self.i2crd1(addr,memloc)&0xff)
      return (val,)
    #elif blen==2:
    #  val= self.i2crd2(addr,memloc)
    #  return ((val&0xff), ((val>>8)&0xff))
    else:
      #print "Error:i2crm: max. 2 bytes available with i2cmainttc (requested:%d)"%blen
      values= []
      if b2: # use (i2cttcpon) 2bytes readings
        for ib in range(blen/2):
          val= self.i2crd2(addr,memloc+ib*2,elog)
          if val==None: 
            values= None
            break
          else:
            values.append(val&0xff)
            values.append((val>>8)&0xff)
        if blen/2*2 != blen:   # last byte
          val= self.i2crd1(addr,memloc+blen-1)
          if val==None: values= None
          else:
            values.append(val&0xff)
      else: # use (i2cttcpon) 1byte readings
        for ib in range(blen):
          val= self.i2crd1(addr,memloc+ib)
          if val==None: values= None
          else:
            values.append(val&0xff)
      return values
  def i2c_read(self,addr,memloc, blen=1):   # used only for SFP.getid_ttcpon
    val= []
    if blen==1:
      v= self.i2crd1(addr,memloc)
      if v==None: v = -1
      else: v= v&0xff
      val.append(v)
    else:
      v= self.i2crd2(addr,memloc)
      if v==None: return (-1,-1)
      val.append(v&0xff)
      val.append((v>>8)&0xff)
    return val
    
class I2Ccore(GenI2C):
  """ Usage:   (Wishbone, rherveille@opencores.com)
  # i2c ops:
  b.i2ci(khz=400)
  b.i2cw(addr, val)  - val: 8bits
  b.i2cwm(addr, memloc, values):
  values= b.i2cr(addr, blen=1):
  values= b.i2crm(addr,memloc, blen=1):

"""
  CP_LO= 0      # Clock Prescale lo-byte
  CP_HI= 1      #                hi-byte
  CTRL= 2       # control
  DATA= 3       # transmit/receive
  CMD_STAT= 4   # Command/Status
  STATUS_BITS= ("IF", "TIP", "R2", "R3", "R4", "AL", "BSY", "RxACK")
  STOP_CMD= 0x40
  READ_CMD= 0x20
  WRITE_CMD= 0x10
  NACK_CMD=0x8
  ACK_CMD=0x0
  def __init__(self, board, i2cslave="i2c", mxs=None, khz=400):
    """ set Prescale regs, 2 options:
    100  -> 100khz clock
    400  -> 400khz clock
    """
    GenI2C.__init__(self, board, i2cslave, mxs)
    self.ur= "%s.userregs"%(self.i2cslave)
    self.i2cmaindbg= False
    self.board.Write_nd(self.ur, 0x00, I2Ccore.CTRL)    # I2C core disabled
    # (32000/5*100khz)-1= 0x3f
    hilo= (32000/(5*khz))-1
    self.board.Write_nd(self.ur, hilo&0xff, I2Ccore.CP_LO)
    self.board.Write_nd(self.ur, hilo>>16, I2Ccore.CP_HI)
    self.board.Write(self.ur, 0x80, I2Ccore.CTRL)    # I2C core enabled
    #print "%s Prescale: %d khz: 0x%x"%(i2cslave,khz,hilo)
  def rawin(self, msg="press any key..."):
    pass
    #raw=raw_input(msg)
  def waitTIP(self):
    #print "1sec sleep..." ; time.sleep(1)
    for ix in range(10):
      stat=self.board.Read(self.ur, I2Ccore.CMD_STAT) #self.dispatch()
      if (stat & 0x2)==0: break   # TIP:0 = transfer complete
    return stat
  def checkACK(self,txt="???"):
    stat= self.waitTIP()
    if ((stat & 0x80)==0) and ((stat & 0x2)==0): 
      #msg= "%s, ACK received (%d loops) status:0x%x"%(txt,ix, stat)
      return stat
    else:
      if txt!="":
        msg= "%s, ACK not received: status:0x%x, sending STOP..."%(txt, stat)
        print msg
      self.board.Write(self.ur, I2Ccore.READ_CMD|I2Ccore.STOP_CMD, I2Ccore.CMD_STAT)
      return None
  def prtSTATUS(self,stat=None):
    #if stat==None: 
    #  stat=self.board.Read("i2c.userregs", CMD_STAT)
    #  self.board.dispatch()
    #strs=""
    #for ix in range(8):
    #  if stat & (1<<ix): strs= strs+ STATUS_BITS[ix] + " "
    #return "STATUS: 0x%x = %s"%(stat, strs)
    lo=self.board.Read(self.ur, self.CP_LO)
    hi=self.board.Read(self.ur, self.CP_HI)
    return "Prescale:0x%x (f:400k, 3f:100k)   "%((hi<<8) | lo) + self.getSTATUS()
  def getSTATUS(self,stat=None):
    if stat==None: 
      stat=self.board.Read(self.ur, I2Ccore.CMD_STAT) #self.dispatch()
    strs=""
    for ix in range(8):
      if stat & (1<<ix): strs= strs+ I2Ccore.STATUS_BITS[ix] + " "
    mxs= self.getMuxes()
    return "%s STATUS: 0x%x = %s MUXs:%s"%(self.i2cslave, stat, strs, mxs)
  def i2ci(self, khz=400):
    """ set Prescale regs, 2 options:
    100  -> 100khz clock
    400  -> 400khz clock
    """
    self.board.Write_nd(self.ur, 0x00, self.CTRL)    # I2C core disabled
    # (32000/5*100khz)-1= 0x3f
    hilo= (32000/(5*khz))-1
    self.board.Write_nd(self.ur, hilo&0xff, self.CP_LO)
    self.board.Write_nd(self.ur, hilo>>16, self.CP_HI)
    self.board.Write(self.ur, 0x80, self.CTRL)    # I2C core enabled
    print "i2c Prescale: %d khz: 0x%x"%(khz,hilo)

  def i2cw(self, addr, value):
    """ value: single value (8bits)
    """
    # WR slave addr (wrbit)
    self.board.Write_nd(self.ur, (addr<<1) | 0 ,I2Ccore.DATA)   # addr + WR2SLAVE
    self.board.Write(self.ur, 0x90, I2Ccore.CMD_STAT)   # STA+ WR
    if not self.checkACK("i2cw STA+WR addr:0x%x"%addr): return None
    # WRdata
    self.board.Write_nd(self.ur, value, I2Ccore.DATA)
    #self.board.Write(self.ur, 0x10, I2Ccore.CMD_STAT)   #  WR
    self.board.Write(self.ur, 0x50, I2Ccore.CMD_STAT)   #  WR+STO
    if not self.checkACK("i2cw DATAsent:0x%x"%value): return None
    # goto WRdata if multiple bytes...
    #STOP done with last byte
    #self.board.Write(self.ur, 0x50, I2Ccore.CMD_STAT)   # WR+STO
    #print "STO sent."
  def i2cwm(self, addr, memloc, values):
    """ addr:  i2c slave addr
    memloc:    memory location
    values:    single value (8bits) or
               list of values, e.g.: [3,67,244] -> 3 data bytes out
    return:    number of bytes acknoledged
    """
    nsent= 0
    #if type(values) == types.ListType:
    if type(values) != types.IntType:
      value= values
    else:
      value= (values,)
    # WR slave addr (wrbit)
    self.board.Write_nd(self.ur, (addr<<1) | 0 ,I2Ccore.DATA)   # addr + WR2SLAVE
    self.board.Write(self.ur, 0x90, I2Ccore.CMD_STAT)   # STA+ WR
    if not self.checkACK("i2cwm STA+WR addr:%x"%addr): return None
    # WR memloc
    self.board.Write_nd(self.ur, memloc, I2Ccore.DATA)   # memloc
    self.board.Write(self.ur, 0x10, I2Ccore.CMD_STAT)   # STA+ WR
    if not self.checkACK("i2cwm WR memloc:0x%x"%memloc): return None
    blen= len(value)
    wrstop= I2Ccore.WRITE_CMD
    for ib in range(blen):
      if ib==(blen-1):   # last byte
        wrstop= I2Ccore.WRITE_CMD | I2Ccore.STOP_CMD
      # WR data
      self.board.Write_nd(self.ur, value[ib], I2Ccore.DATA)
      self.board.Write(self.ur, wrstop, I2Ccore.CMD_STAT)   #  WR
      if not self.checkACK("i2cwm DATAsent %d:0x%x"%\
        ((memloc+ib),value[ib])): return None
      nsent= nsent+1
    #STOP done with last byte
    #self.board.Write(self.ur, 0x50, I2Ccore.CMD_STAT)   # WR+STO
    if nsent!= len(value): raise "i2cwm: %d but only %d sent"%(len(value), nsent)
    return nsent
  def i2cr(self, addr, blen=1):
    """ single reg. read (multiple readings if blen>1)
    return: list of values or 
            single value in case blen==1
            None in case of error 
    """
    # WR slave addr (rdbit)
    readval= []
    self.board.Write_nd(self.ur, (addr<<1) | 1 ,I2Ccore.DATA)   # addr+ READSLAVE
    self.board.Write(self.ur, 0x90, I2Ccore.CMD_STAT)        # STA+WR
    if not self.checkACK("i2cr STA"): return None
    for ib in range(blen):
      if ib==(blen-1):   # last byte
        acknak= I2Ccore.NACK_CMD | I2Ccore.STOP_CMD
        acknaks= "NACK+STOP"
      else:
         acknak= I2Ccore.ACK_CMD
         acknaks= "ACK"
      #self.rawin("press a key. Next: read data + %s..."%acknaks)
      self.board.Write(self.ur, I2Ccore.READ_CMD|acknak, I2Ccore.CMD_STAT)
      readval.append(self.board.Read(self.ur, I2Ccore.DATA))
      #print "Byte:%d DATAreceived:0x%x"%(ib,readval)
      #
    if len(readval)==1: return readval[0]
    if len(readval)==0: return None
    return readval
  def i2crm(self,addr,memloc, blen=1, si534x=False, elog=True):
    """ addr:  i2c slave addr
    memloc:    memory location from where to read
    blen:      number of required readings
    si534x:    True only for Si534x (see page 52 in Si534x manual)
    return:    list of values, e.g.: [3,67,244] -> 3 data bytes out
               None in case of error
    """
    # START
    self.board.Write_nd(self.ur, (addr<<1) | 0 ,I2Ccore.DATA)   # addr+WRITESLAVE
    self.board.Write(self.ur, 0x90, I2Ccore.CMD_STAT)        # STA+WR
    if elog:
      txtsta="i2crm STA"
    else:
      txtsta=""
    if not self.checkACK(txtsta): return None
    # memloc
    #self.rawin("press a key. Next: write memloc...")
    self.board.Write_nd(self.ur, memloc, I2Ccore.DATA)       # mem. location
    if si534x:
      cmdbyte= self.WRITE_CMD | self.STOP_CMD        # WR+STO
    else:
      cmdbyte= self.WRITE_CMD        # WR
    self.board.Write(self.ur, cmdbyte, I2Ccore.CMD_STAT)        # WR+ [STO]
    if not self.checkACK("i2crm memloc"): return None
    # START (repeated)
    #self.rawin("press a key. Next: repeated start...")
    self.board.Write_nd(self.ur, (addr<<1) | 1 ,I2Ccore.DATA)   # addr+RDSLAVE
    #self.board.Write(self.ur, 0xa0, I2Ccore.CMD_STAT)        # repeated STA+RD
    self.board.Write(self.ur, 0x90, I2Ccore.CMD_STAT)        # repeated STA+WR
    if not self.checkACK("i2crm repSTA"): return None
    readval= []
    for ib in range(blen):
      if ib==(blen-1):   # last byte
        acknak= I2Ccore.NACK_CMD | I2Ccore.STOP_CMD
        acknaks= "NACK+STOP"
      else:
        acknak= I2Ccore.ACK_CMD
        acknaks= "ACK"
      #self.rawin("press a key. Next: read data + %s..."%acknaks)
      self.board.Write(self.ur, I2Ccore.READ_CMD|acknak, I2Ccore.CMD_STAT)
      val=self.board.Read(self.ur, I2Ccore.DATA)
      readval.append(val)
      #print "Byte:%d DATAreceived:0x%x"%((memloc+ib),val)
    return readval
class GenDev:
  def __init__(self, conn, devname, addr, mux):
    self.b= conn    # bus (an I2Ccore/I2Ctpcore (SPIcore?) instance)
    self.devname= devname
    # mux: 1 pair for each mux. addr 0 reserved for virtmux in fpga
    self.mux= mux        # (muxaddr1,chan1, muxaddr2, chan2,...)
    self.addr= addr      # i2c address
    self.lastdevid= "neverchecked"
  def setmux(self):
    """ Has to be invoked in each method accesing device's registers
mux: (addr, channel)    e.g.: (0x74, 0x20)
     None  no mux
     addr:0 -reserved for virtmux (virtual mux in fpga)
     (a1,c1, a2,c2) -cascaded/parallel muxes
     cascaded: (a1,c1) programmed before (a2,c2)!
     parallel: the order of programming doesn't matter
    
setmux is called:
-in each method touching hw regs
Note about i2cmainttc mux (1..9:olt1..9 10:onu 11:ipbus):
Is set when needed (i.e. not here)

"""
    if self.mux: 
      for ix in range(0,len(self.mux),2):
        ma= self.mux[ix]
        if ma==0:   # mux implemented in ctpltu fpga
          self.b.board.Write("ctrl.fmcsfpsel", self.mux[ix+1])
        else:
          self.b.i2cw(ma, self.mux[ix+1])   # set mux
          #print "setmux: 0x%x: 0x%x"%(ma, self.mux[ix+1])
    #mmux= self.b.board.Read("i2cmainttc.mux")  -not available in tst_logic!
    #print "i2cmainttc.mux:",mmux
  def getid(self):   # overwritten by device's method
    """ should return None or string composed from something read from device
"""
    return "getid not implemented"
  def excmd(self,cmd):
    #self.setmux()   -not needed here, present in each method!
    if len(cmd)<=1:
      self.prthelp()
      return
    self.myexcmd(cmd)

class Si534x(GenDev):   # ctp: si5345  kcu:si5344
  Default_hpy_before= "ok_ltu_logic.hpy"
  Default_hpy_c00= "pll_in0_global_c00.hpy"   # from fi 0c_00
  Default_hpy_mft= "pll_in0_global_c00_itsmft.hpy"
  Default_hpy_ctp= "pll_in1_stdalone_c00.hpy"
  def __init__(self, conn, devname="si5345", addr= 0x68, mux=None, model="5345"):
    GenDev.__init__(self, conn, devname,addr, mux)
    self.model= model
    self.notchecked=(0x0b24, 0x0b25, 0x0540, 0x035a, 0x035c, 0x0514, 0x001c)
    self.Default_hpy= self.Default_hpy_before
    fw= self.b.board.fwinfo
    if fw:
      fwinfotyp= fw>>24
      fwinfover= fw&0xffffff
      if fwinfotyp == self.b.board.FW_LTU_LOGIC:
        if fwinfover >= 0xc00:
          self.Default_hpy= self.Default_hpy_c00
      elif fwinfotyp == self.b.board.FW_CTP_LOGIC:
        self.Default_hpy= self.Default_hpy_ctp
      elif fwinfotyp == self.b.board.FW_LTU_ITSMFT_LOGIC:
        self.Default_hpy= self.Default_hpy_mft
      else:
        self.Default_hpy= self.Default_hpy_ctp     # from 26.9.2019
        #print "Error: Unknown firmware 0x%x"%fwinfotyp
    #print "si534x:", "fi 0x%x ver:"%fwinfotyp, hex(fwinfover), self.Default_hpy
    #print "si534x:", hex(self.b.board.fwinfo), "Default_hpy:",self.Default_hpy
    self.alreadychecked= False
  def getid(self):   # p.79
    self.setmux()
    self.b.i2cwm(self.addr, 1, 0)   # select page0
    v= self.b.i2crm(self.addr, 0x03, 1)
    if v==None or len(v)<1: return None
    base= v[0]
    v= self.b.i2crm(self.addr, 0x02, 1) ; base= (base<<8) | v[0]
    pn= "%x"%base
    return pn
  def wreg(self, page,reg,value):
    self.setmux()
    self.b.i2cwm(self.addr,1,page)
    self.b.i2cwm(self.addr,reg,value)
  def rreg(self, page,reg, newway=False):
    self.setmux()
    if page>=0:
      self.b.i2cwm(self.addr,1,page)
    if newway:
      if page<0:
        self.b.i2cw(self.addr, 1)   # PAGE
        cpage= self.b.i2cr(self.addr)
        print "Using current (0x%x) page"%(cpage)
      v= self.b.i2crm(self.addr, reg, 1, True)   # 1 byte
      if v!= None and len(v)>0: v= v[0]
      #self.b.i2cw(self.addr, reg)
      #v= self.b.i2cr(self.addr)
    else:
      if page<0:
        cpage= self.b.i2crm(self.addr, 1, 1)   # 1 byte from PAGE
        if cpage!= None and len(cpage)>0: cpage= cpage[0]
        print "Using current (0x%x) page"%(cpage)
      v= self.b.i2crm(self.addr, reg, 1)   # 1 byte
      if v!= None and len(v)>0: v= v[0]
    return v
  def getregs(self,fname):
    #fpath= os.path.join(os.getenv("CTP3_CONFIG_DIRECTORY"), fname)
    fpath= os.path.join(os.getenv("CTP3_ROOT"), "boards", "cfg", fname)
    print "Si5343x:%s, opening %s..."%(self.devname, fpath)
    fd= open(fpath,"r") ; regstx= fd.read() ; fd.close()
    regstx= regstx.replace('{','(')
    regstx= regstx.replace('}',')')
    regs= eval(regstx)
    #for ix in range(4):
    #  print ix,":","0x%x: 0x%x"%(regs[ix][0], regs[ix][1])
    return regs
  def setname(self, sgname):
    if sgname==None: name= self.Default_hpy
    elif (sgname=='g') or (sgname=='global'): sgname= "global"
    elif (sgname=='s') or (sgname== "stdalone"): sgname= "stdalone"
    else:
      print "Error calling i2c.si5345.setname(%s)"%sgname
      return None
    fwi= self.b.board.fwinfo
    #print "i2c.py setname1:", hex(fwi), name
    fwinfotyp= self.b.board.fwinfo>>24
    if fwinfotyp == self.b.board.FW_LTU_LOGIC:
      if fwi >= 0x1100:   # cdc: always pll_in1
        name= "pll_in1_stdalone_c00.hpy"
      elif fwi >= 0xc00:
        if sgname == "stdalone":
          name= "pll_in1_stdalone_c00.hpy"
        elif sgname == "global":
          name= "pll_in0_global_c00.hpy"
        else:
          pass
      else:
        if sgname == "stdalone":
          name= "pll_in1_stdalone.hpy"
        elif sgname == "global":
          name= "pll_in0_global.hpy"
        else:
          pass
    elif fwinfotyp == self.b.board.FW_CTP_LOGIC:
      if sgname == "stdalone":
        name= "pll_in1_stdalone_c00.hpy"
      elif sgname == "global":
        print "ERROR? -CTP board in global mode"
        name= "pll_in0_global_c00.hpy"
      else:
        pass
    elif fwinfotyp == self.b.board.FW_LTU_ITSMFT_LOGIC:
      if fwi >= 0x1100:   # cdc: always pll_in1
        name= "pll_in1_stdalone_c00_itsmft.hpy"
      else:
        if sgname == "stdalone":
          name= "pll_in1_stdalone_c00_itsmft.hpy"
        elif sgname == "global":
          name= "pll_in0_global_c00_itsmft.hpy"
        else:
          pass
    else:
      name= self.Default_hpy
    #print "i2c.py setname:", hex(fwi), name
    return name
  def load(self, name=None):   # unconditional load
    """ name: "stdalone" or "global"
"""
    self.setmux()
    name= self.setname(name)
    regs= self.getregs(name)
    prevpage= 0xff
    for ix in range(len(regs)):
      rd= regs[ix]
      curpage= rd[0]>>8
      if curpage!=prevpage:
        #print "====== page: 0x%x)"%curpage
        self.b.i2cwm(self.addr, 1, curpage)
      memloc= rd[0]& 0xff
      #print "i2cwm(self.addr, 0x%x, 0x%x)"%(memloc,rd[1])
      self.b.i2cwm(self.addr, memloc, rd[1])
      #self.b.i2cwm(self.addr, memloc, rd[1])   duplicate line (commented 22.8.)
      if (curpage==0x0b) and (memloc==0x25) and (rd[1]==0):
        print "waiting 300ms after preamble..."
        time.sleep(0.3)
      prevpage= curpage
    print colors.color("Si%s: %s loaded."%(self.model, name), "fg_red")
    time.sleep(0.045)
    # wait 45ms:
    # The spec you are looking for is referred to as "Initial Start-up Time" in
    # the Si5345 datasheet (page 32, Table 5.8).
    # https://www.silabs.com/documents/public/data-sheets/Si5345-44-42-D-DataSheet.pdf
    #
    # set LED1:stda or LED2: global -> Session.odevs["LEDS"]
    # self.b.board... but .odevs is in Session object:
    # - move odevs (+ all devs?) to ATBoard and
    # ATBoard in atbwrap.py instead of ipbuswrap.Board ?
    self.clearSticky()
    self._oltreset()
    self.checkLOSL()
  def checkit(self, pgml):   # rc: T if to be checked, F: do not check this byte
    if pgml in self.notchecked: return False
    return True
  def check(self, name=None, stdout=True):   # rc: T if ok, F if check unsuccesfull
    """ name: "stdalone" or "global"
"""
    self.setmux()
    name= self.setname(name)
    regs= self.getregs(name)
    prevpage= 0xff
    rc= True
    for ix in range(len(regs)):
      rd= regs[ix]
      curpage= rd[0]>>8
      if curpage!=prevpage:
        if stdout: print "====== page: 0x%x)"%curpage
        self.b.i2cwm(self.addr, 1, curpage)
      memloc= rd[0]& 0xff
      pgml= (curpage<<8) | memloc
      #print "i2cwm(self.addr, 0x%x, 0x%x)"%(memloc,rd[1])
      v= self.b.i2crm(self.addr, memloc, 1)
      if v==None:
        print "error reading PLL cfg mem"
        return None
      if (v[0] != rd[1]) and self.checkit(pgml):
        #print "0x%2.2x%2.2x: 0x%x  exp:0x%x"%(curpage, memloc, v[0], rd[1])
        if stdout: print "0x%4.4x: 0x%x  exp:0x%x"%(pgml, v[0], rd[1])
        rc= False
      if (curpage==3) and (memloc==0x5c):
        if stdout: print "0x%2.2x%2.2x: 0x%x  <--- phase"%(curpage, memloc, v[0])
      prevpage= curpage
    #self.checkLOSL()
    if rc:
      print "agrees with PLL"
      print "Checking (+ clearing if not ok) LOL+LOS sticky bits..."
      rc2,msg= self.cclosl()
      print msg
    else:
      print "does not agree with PLL"
    return rc
  def checkandload(self, name=None):
    if self.alreadychecked == True:
      print "Si534x checkandload: already checked not done again."
      print "Checking (+ clearing if not ok) LOL+LOS sticky bits..."
      rc2,msg= self.cclosl()
      print msg
      return
    else:
      print "Si534x checkandload:", self.b.board.fwinfo," ..."
    if self.check(name, False): 
      #print "PLL config checked, ok."
      print colors.color("Si5345 config: %s, ok"%name, "fg_green")
      self.alreadychecked= True   # i.e. check more after 'load'
    else:
      print "PLL config not ok, reloading..."
      self.load(name)
  def checkgs(self, mode, load,checkbefore=True):
    """ mode: 'g'(global) or 's' (stdalone)
load: True -load if necessary, False: only print warning
"""
    if (mode != 'g') and (mode != 's'):
      print colors.color("Bad mode in Si534x/checksgs: %s"%mode, "fg_red")
      return
    if checkbefore:
      chk= self.check(mode, False)
      #print "PLL config checked, ok."
      if chk==True:
        print colors.color("Si5345: config ok, mode: %s"%mode, "fg_green")
      elif chk==False:
        if load:
          print "Reloading PLL..."
          self.load(mode)
      else:
        pass   # errror readng PLL memory
    else:
      print colors.color("Si5345: bad config for mode %s"%mode, "fg_red")
      if load:
        print "Reloading PLL..."
        self.load(mode)
  def checkLOSL(self):
    self.setmux()
    pn= None
    self.b.i2cwm(self.addr, 1, 0)   # select page0
    v= self.b.i2crm(self.addr, 0x03, 1)
    if len(v)<1: return pn
    base= v[0]
    v= self.b.i2crm(self.addr, 0x02, 1) ; base= (base<<8) | v[0]
    pn= "%x"%base
    v= self.b.i2crm(self.addr, 0x0d, 1) ; los= v[0]&0xf           # LOS     p24
    v= self.b.i2crm(self.addr, 0x12, 1) ; lossticky= v[0]&0xf     # LOS_FLG p24
    v= self.b.i2crm(self.addr, 0x0e, 1) ; lol= (v[0]&2)>>1        # LOL     p28
    v= self.b.i2crm(self.addr, 0x13, 1) ; lolsticky= (v[0]&2)>>1  # LOL_FLG p28
    v= self.b.i2crm(self.addr, 0x2c, 1) ; los_en= v[0]&0xf        # LOS_EN
    self.b.i2cwm(self.addr, 1, 0x09)   # select page0
    v= self.b.i2crm(self.addr, 0x49, 1) ; in_en= v[0]&0xf
    print "Enabled inputs IN3..0:0x%x "%(in_en)
    print "Enabled LOS    IN3..0:0x%x "%(los_en)
    print "Sticky  LOS    IN3..0:0x%x "%(lossticky)
    print "Base pn:0x%8x Loss of Signal IN3..0:0x%x  Loss of Lock(1 bit)/sticky:%d/%d"%(base,los,lol, lolsticky)
    return pn
  def _waitready(self, wmsk=0x6):
    nloops= 0 ; slt= 0.5
    while nloops<20:
      #val= self.read_ctrl_reg(self.Astatus)
      val= self.b.board.Read("ttcponolt_u1.uregs", 3)
      if (val & wmsk) == wmsk:
        print "_waitready ok, (%.2f s): waitmask: 0x%x status:0x%x"%(slt*nloops, wmsk, val)  
        return
      time.sleep(slt)
      nloops= nloops+1
    print "_waitready timeout (%.2f s): waitmask: 0x%x status:0x%x"%(slt*nloops, wmsk, val)  
  def _oltreset(self, wmsk=0x6, msg="OLT reset after PLL LOS/L flags cleared done."):   # 0x40:rxReady 0x20: txReady
    if self.b.board.nodExists("ttcponolt_u1"):
      stat= self.b.board.Read("ttcponolt_u1.uregs", 0)
      stat = stat | 1
      self.b.board.Write("ttcponolt_u1.uregs", stat, 0)
      stat = stat ^ 1
      self.b.board.Write("ttcponolt_u1.uregs", stat, 0)
      print msg
      self._waitready(wmsk)
    else:
      print "OLT reset not done (OLT not available)"
  def clearSticky(self):
    #time.sleep(1)   # 1 s enough after reset?
    self.b.i2cwm(self.addr,1, 0)  # page0
    self.b.i2cwm(self.addr,0x12, 0)     # LOS_FLG
    self.b.i2cwm(self.addr,0x13, 0)     # LOL_FLG
    #if self.b.board.isfiwa((self.b.board.FW_CTP_LOGIC, self.b.board.FW_LTU_LOGIC)):
    #  # always OLTreset after LOL:
    #  self._oltreset()
  def cclosl(self):   # check+clear losl bits for s/g modes
    '''returned: rc,msg: error, a message about ONU/input clock
       rc:0 ok
          4 LOL or LOS detected
          8 internal error
'''
    self.setmux()
    self.b.i2cwm(self.addr, 1, 0)   # select page0
    v= self.b.i2crm(self.addr, 0x2c, 1) ; los_en= v[0]&0xf        # LOS_EN
    v= self.b.i2crm(self.addr, 0x12, 1) ; lossticky= v[0]&0xf  # LOS_FLG p24
    v= self.b.i2crm(self.addr, 0x13, 1) ; lolsticky= (v[0]&2)>>1  # LOL_FLG p28
    self.b.i2cwm(self.addr, 1, 0x09)   # select page0
    v= self.b.i2crm(self.addr, 0x49, 1) ; in_en= v[0]&0xf         # enabled inputs
    if (los_en== in_en) and (los_en==0xa):    # stdalone: in1 1010 (pllin1)
      #msg1= "Front panel input clock (pllin1:stdalone)"
      msg1= "FP/PLL clock (pllin1:stdalone)"
      chanbit= 0x2  
    elif (los_en== in_en) and (los_en==0x9):    # global:in0 NEVER with CDC (fi>v10_08)
      msg1= "ONU clock (pllin0:global)"
      chanbit= 1
    else:
      msg= "Error: Bad Si5345 config (not stdalone neither global)."
      return 8,msg   # error
    signalok= False
    rc= 0
    if lolsticky==0:
      if (chanbit & lossticky) == 1:
        msg= "Loss of %s signal detected."%msg1
      else:
        msg= "%s: signal ok."%msg1
        signalok= True
    else:
      if (chanbit & lossticky) == 1:
        msg= "LOL+LOS detected for %s."%msg1
      else:
        msg= "Loss of Lock (but signal was ok) detected for %s."%msg1
    if not signalok:
      msg= msg+ " Clearing LOL+LOS sticky bits..."
      self.clearSticky()
      rc= 4
    return rc,msg
  def get_status(self):     # used in init_onu_fmc.py
    self.setmux()
    self.b.i2cwm(self.addr, 1, 0)   # select page0
    v= self.b.i2crm(self.addr, 0x0d, 2)
    los= (v[0]&0x2)>>1   # IN1 LOS
    lol= (v[1]&0x2)>>1   # LOL
    #print "Si5345 LOS LOL:0x%x 0x%x"%(v[0], v[1])
    los=lol= v[0]&1   # IN0 LOS
    return [1, los, lol]
  def SetSkew(self,pll_out_skew, xx=1):
    """ Nxx_DELAY (xx: 0..4) 5 registers available, each 2x 16bits
0x0359 0x35a: N0_DELAY          
0x035b 0x35c: N1_DELAY  default OUT7 120.2355 MHz (goodphase)
0x035d 0x35e: N2_DELAY  default OUT6 240.471 MHz (onu2oltgoodphase)
...
"""
    ndeladdr= 0x59 + xx*2
    #self.setmux() assuming mux is set, i.e.:  b.i2cw(0x74, 8) for ltukcu
    self.b.i2cwm(self.addr,1, 3)  # page3
    self.b.i2cwm(self.addr,ndeladdr, 0)            # register 0x0359 - N0_DELAY low bits
    self.b.i2cwm(self.addr,ndeladdr+1, pll_out_skew) # upper bits
    #self.b.i2cwm(self.addr,0x5b, 0)            # register 0x035b - N1_DELAY low bits
    #self.b.i2cwm(self.addr,0x5c, pll_out_skew) # upper bits
    #time.sleep(0.1)   # removed 6.12.2017 (onu init,fullcal working without this sleep)
    self.b.i2cwm(self.addr,1,0)   # page0
    self.b.i2cwm(self.addr,0x1c,1)   #soft reset
    return 0
  def ReadDelays(self):
    self.setmux()
    self.b.i2cwm(self.addr,1, 3)  # page3
    #v= self.b.i2crm(self.addr, 0x5a, 7)   # 0x5a* b c* d e* f 0x60*
    #print "N[0..3]_DELAY: 0x%x 0x%x 0x%x 0x%x"%(v[0],v[2], v[4], v[6])
    v= self.b.i2crm(self.addr, 0x59, 10)   # 0x5a* b c* d e* f 0x60*
    msg= "N[0..4]_DELAY: 0x"
    for ix in range(5):
      msg= msg + "(%x %x) "%(v[ix*2], v[ix*2+1])
    print msg
  def Sphase(self,phase, xx=1):   # invoked also from ublaze (init_onu_fmc.py)
    self.setmux()                 # and init_onu_mlinkolt_cascaded_fmc.py
    self.SetSkew(phase,xx)
  def Gphase(self, xx=1):   # invoked also from ublaze (init_onu_fmc.py)
    self.setmux()                 # and init_onu_mlinkolt_cascaded_fmc.py
    self.b.i2cwm(self.addr,1, 3)  # page3
    v= self.b.i2crm(self.addr, 0x59, 10)   # 0x5a* b c* d e* f 0x60*
    phase= (v[xx*2]<<16)+ v[xx*2+1]
    return phase
  def Fphase(self):
    # check if ONU present (SFP9)
    onudev= self.b.findDevice("SFP9")
    #if (onudev==None) and (self.b.board.bname!="proto_X") and \
    #  (self.b.board.bname!="ltukcu") and (self.b.board.bname!="proto_2"):
    if onudev==None:
      print "ONU (SFP9) not connected..."
      return
    self.setmux()
    minp=maxp=prevphase= -1
    for skew in range(0x80):   # 0.. 0x7f
      self.SetSkew(skew)
      # reset onu: soft reset done in SetSkew
      time.sleep(1)   # wailt PLL locked, at least 600ms in ZDM
      status= self.b.board.Read("ttcpononu_u.uregs",3)
      phase= (status>>1)&1
      if (phase==1) and (minp==-1): minp= phase   # 1st 1
      if (phase==0) and (minp!=-1): maxp= phase   # 1st 0 after 1
      if phase != prevphase:
        if phase==1 or skew<3:
          print "%d: 0x%x %d..."%(skew,status,phase)
    print "good phase: (%d+%d)/2= %d ?"%(minp,maxp,(minp+maxp)/2)
  def hwreset(self):
    self.setmux()
    #self.b.i2cwm(self.addr,1,0)   # page0
    #self.b.i2cwm(self.addr,0x1e,1)   #hard reset
    fwinfo= self.b.board.Read("ctrl.fwinfo")
    fwinfo= self.b.board.fwinfo
    if ((fwinfo >> 24) == 0x4) or ((fwinfo >> 24) == 0xab):   
      # 0x4:kcu105 (still 0xab) -> see boards/readme
      # let's use insted RST pin on si5344 controlled by fy:
      self.b.board.Write("i2c.userregs", 1, 7)
      self.b.board.Write("i2c.userregs", 0, 7)
      print "PLL_RST: 1->0 done"
    else:
      self.b.i2cwm(self.addr,1, 0)  # page0
      self.b.i2cwm(self.addr,0x1e, 2)     # 1:hard rst
      #1 causes hard reset. The same as power up except 
      #that the serial port access is not held at reset. This 
      #does not self-clear, so after setting the bit it must be 
      #cleared.
      self.b.i2cwm(self.addr,0x1e, 0)     # 0
      print "hard rst done."  
  def excmd(self,cmd):
    self.setmux()
    if len(cmd)<=1:
      self.prthelp()
      return
    if cmd[1]=='id':
      print "id:", self.getid()
    elif cmd[1]=='loadpll':
      name=None
      if len(cmd)>2: name= cmd[2]
      self.load(name)
      #pll_shift= 0x3f   # 0..0x7f
      #rc= self.SetSkew(b,pll_shift)
      #print "self.SetSkew(%d) rc:%d"%(pll_shift, rc)
    elif cmd[1]=='checkpll':
      name=None
      if len(cmd)>2: name= cmd[2]
      self.check(name)
    elif cmd[1]=='losl':
      self.checkLOSL()
    elif cmd[1]=='loslclear':
      self.clearSticky()
    elif cmd[1]=='cclosl':
      print self.cclosl()
    elif cmd[1]=='wreg':
      page= eval(cmd[2])
      reg = eval(cmd[3])
      val = eval(cmd[4])
      print "writing 0x%x 0x%x:0x%x..."%(page, reg, val)
      self.wreg(page,reg,val)
    elif cmd[1]=='rreg':
      if cmd[2]=='c': page= -1   # use current  page
      else: page= eval(cmd[2])
      reg = eval(cmd[3])
      v= self.rreg(page,reg)
      if v!=None: print "read 0x%x 0x%x:0x%x"%(page, reg, v)
    elif cmd[1]=='nreg':
      print "not needed (used for debug only)"
      #page= eval(cmd[2])
      #reg = eval(cmd[3])
      #v= self.rreg(page,reg, True)
      #if v!=None: print "readNEW WAY 0x%x 0x%x:0x%x"%(page, reg, v)
    elif cmd[1]=='phases':
      #self.b.connected[0].ReadDelays()
      self.ReadDelays()
    elif cmd[1]=='fphase':
      #self.b.connected[0].Fphase()
      self.Fphase()
    elif cmd[1]=='sphase':
      #self.b.connected[0].Sphase(eval(cmd[2]))
      self.Sphase(eval(cmd[2]))
    elif cmd[1]=='hwreset':
      self.hwreset()
    else:
      self.prthelp()
  def prthelp(self):
    print """
%s id               -print id (5345 or 5344)
    loadpll [fn]    -init Si534x + read LOS/LOL bits (default: 
                     $CTP3_ROOT/boards/cfg/%s)
    checkpll [fn]   -read+compare with expected values
                     fn -data file to be loaded from file ~/WORK/fn
    losl            -check Loss of Signal/Lock actual and sticky bits
    loslclear       -clear sticky bits
    cclosl          -check Loss of Lock/Signal considering s/g mode + clear sticky bits if needed
    phases          -read+show 4 deviders N[0..3]_DELAY
    fphase          -find phase good of ONU
    sphase N        -set phase N (0..127)
    hwreset         -reset (i.e. falling edge 1->0 on Si5344 pin PLL_RST)
    rreg page addr  -read register
                     page: c -> use current page
    wreg page addr value   -write register

"""%(self.devname, self.Default_hpy)
class SFP(GenDev):
  def __init__(self, conn, devname, purpose, addr= 0x50, mux=None):
    GenDev.__init__(self, conn, devname,addr, mux)
    # devname: SFPX
    # purpose: Gbt instance of corresponding gbt link
    #          None SFP (ipbus) or unknown or not connected
    #          'gbt' -SFP in, but .xml does not contain corresponding node
    #          'olt' -> not implemented yet
    
    self.purpose= None
    ident= self.getid()   #("all")
    #print "SFP ident:", ident, devname
    if devname=="SFP":
      self.purpose= None   # ipbus SFP
      return
    if ident:
      from atb.gbt import Gbt
      #print "SFP.getid:", devname, ident
      #if purpose[:3]=="gbt":
      if (ident[:4]=="AFBR") or (ident[:4]=="SFP-") or (ident[:4]=="AFCT"):
        # normal, FS, or special ITS gbt
        #gslot= int(purpose[3:])
        for gbtnam in self.b.board.fwsfps:
          if self.b.board.fwsfps[gbtnam] == devname:
            gslot= gbtnam[3:]
            purpose= Gbt(self.b, int(gslot[0]))
            if purpose.ok:
              self.purpose= purpose
            else:
              self.purpose= "gbt"   # i.e. gbt but .xml is not providing "gbtX" top node
            break
      elif ident[:4]=="SOGX":
        self.purpose= "olt"   # todo: olt object instead
      else:
        print "Unknown SFP: "+ident
        self.purpose= None
  def intcalibration(self):    # mux is set already!
    vals92= self.b.i2crm(self.addr, 92, 1, elog=False)[0]
    #if (vals[92] & (1<<6)) and (vals[92] & (1<<5)) and (vals[92] & (1<<2)==0):
    if (vals92 & (1<<6)) and (vals92 & (1<<5)) and (vals92 & (1<<2)==0): return(True)
    return(False)
  def gettemp(self):
    self.setmux()
    if self.intcalibration():
      # internal calibration, t[0]:MSB t[1]: LSB (table 9-11)
      t2= self.b.i2crm(self.addr+1, 96, 2)
      sign= (0x80& t2[0]) >> 7
      h= 0x7f& t2[0]
      l= t2[1]
      if sign==0:
        temp= h+l/256.
      else:
        temp= -1*((~h+1)+(~l)/256.)
      return temp
    else:
      return None
  def getmvolts(self):
    self.setmux()
    if self.intcalibration():
      t2= self.b.i2crm(self.addr+1, 98, 2)   # MSB LSB 1:100uV
      volts= (((t2[0]&0xff)<<8) + (t2[1]&0xff))/10.   # in mV
      return volts
    else:
      return None
  def getrxop(self):   # p. 29 and table 9.7.  on page 35
    # 104: rx power   102: txpower
    self.setmux()
    if self.intcalibration():
      tout= self.b.i2crm(self.addr+1, 102, 2)
      mwsout= (((tout[0]&0xff)<<8) + (tout[1]&0xff)) #*0.0001   #
      t2= self.b.i2crm(self.addr+1, 104, 2)   # MSB LSB 1:0.1uW
      mws= (((t2[0]&0xff)<<8) + (t2[1]&0xff)) #*0.0001   # in mW
      return mws,mwsout
    else:
      return None,None
  def getlastdevid(self,partnumber,sn, intcalib):
      temp= "None"; mvolts= "None" ; mws= "None" ; mwso= "None"
      tempf= self.gettemp() ; mvoltsf= self.getmvolts() ; mwsf,mwsfo= self.getrxop()
      if tempf: temp="%.2fC"%tempf
      if mvoltsf: mvolts="%5.1fmV"%mvoltsf
      if mwsf:
        mws="%5.4fmW"%mwsf
        if intcalib:
          if(partnumber=='LTF7222-BC' or partnumber=='SOGX6292-PSGB'):			
            mwsfo = 10*math.log10((mwsfo * 0.2 * 1e-6)/1e-3) # UNIT: dBm
          else:
            mwsfo = 10*math.log10((mwsfo * 0.1 * 1e-6)/1e-3) # UNIT: dBm	
          mwso="%5.3fdBm"%(mwsfo)
        else:
          mwso="%5.4fmW"%mwsfo
      self.lastdevid= "%s / %s"%(partnumber,sn)
      return "%s / %s \t%s %s %10s %10s"%(partnumber,sn, temp, mvolts, mwso, mws) 
  def getid(self, prt="temp", force=False): 
    """ prt: 
      "" -return partnumber / serial number
      "all" -as "" + print debug info to stdout
      "temp"- return partnumber / serial number temp
    rc: returns None if id cannot be obtained or:
         AFBR-709DMZ / AA161131JRY    47.28C            -gbt
         SFP-10GER-31 / C1911244381   37.48C            -FS gbt
         AFCT-57D3ANMZ / AD17415005K  31.7C             -ITS special gbt
         ABCU-5740ARZ / AGC17485406N  None              -ipbus
         SOGX6292-PSGB / 6586118120002626        40.31" -olt

    Note about force: this parameter used only with SFP now (July2019), its usage
    should be extended to all devices:
    self.lastdevid: "neverchecked", "anystring": last id,  None: not connected (checked at least once)
    force:False:  return self.lastdevid if different from "nc" (not connected)
    force:True:  - repeat getid() algorithm, set self.lastdevid, return None or string
    """
    if force== False:
      if self.lastdevid != "neverchecked": return self.lastdevid
    self.setmux()
    # assuming mux is set. RC: None or PartNumber
    self.partnumber= None
    #vals= self.sfp_read(self.addr, "SFP A0", 0, 127)
    vals= self.b.i2crm(self.addr, 0, 128, elog=False)   # False: do not report error if SFP missing
    if vals==None: 
      self.lastdevid= None
      return self.lastdevid
    if len(vals)<(127-0): 
      self.lastdevid= None
      return self.lastdevid
    # vals[40:56] -> see '7. Vendor fields' in SFF-8472.pdf 
    self.vendor= string.strip(string.join(map(chr,vals[20:36]),""))
    self.partnumber= string.strip(string.join(map(chr,vals[40:56]),""))
    self.sernumber= string.strip(string.join(map(chr,vals[68:84]),""))
    intcalib= False
    if vals[92] & (1<<6):   # see TABLE 8-5 p.26
      if vals[92] & (1<<5):
        intcalib= True
    if prt=="all":
      print "Vendor     :", self.vendor
      print "Vendor OUI : 0x%x"%((vals[37]<<16) | (vals[38]<<8) | vals[39])
      print "Part number:",self.partnumber
      print "Revision   :",string.join(map(chr,vals[56:60]),"")
      print "Vendor SN  :",self.sernumber
      print "YYMMDD date:",string.join(map(chr,vals[84:92]),"")
      print "Diag. monitoring type, vals[92]:",hex(vals[92])
      if vals[92] & (1<<6):   # see TABLE 8-5 p.26
        if vals[92] & (1<<5):
          print "Internally calibrated"
        if vals[92] & (1<<4):
          print "Externally calibrated (A/D counts..., not implemented)"
        if vals[92] & (1<<2):
          print "0x51(A2) access: not implemented (see p.25)"
        else:
          print "0x51(A2) access: ok"
        # temperature:
        if (vals[92] & (1<<5)) and (vals[92] & (1<<2)==0):
          # p.35 Table 9.7
          temp2= self.b.i2crm(self.addr+1, 96, 2)   #, b2=True)
          mvolts_msb_lsb= self.b.i2crm(self.addr+1, 98, 2)
          rxop= self.b.i2crm(self.addr+1, 104, 2)
          print "temp2:", temp2, "mvolts:", mvolts_msb_lsb,"rxop:",rxop 
          # SHL      S:0x80   H:0x7f   L:0xff     see p.29
          # S: 0:+ 1:-   +: t= H+L/256.
          # S: 0:+ 1:-   -: t= (~H + 1) + (~L)/256.
          #if temp2:
          #  print "Temperature high low:", map(hex,temp2)
        else:
          print "temp2: not available"
      else:
        print "Digital diagnostic monitoring not available (bit 92.7 not set)"
      # table 9-11: 96..109
      hexdump.hd_dump(vals, 92, 109)
      if vals[92] & (1<<3):
        print "Bit 92.3 on -> pm: average input optical power AVG"
      else:
        print "Bit 92.3 off-> pm: OMA"
      if vals[92] & (1<<2):
        print "Bit 92.2 on i.e. special A2 addressing!"
      print "Functionality implemented (see SFF-8472.pdf table 8-8):0x%x (0x3: Rev. 12.2)"%vals[94]
      #sfp_read(b, 0x51, "SFP A2 (bit 92.2 not considered)", 127, 135)
      rc= self.getlastdevid(self.partnumber,self.sernumber, intcalib)
    elif prt=="temp":
      rc= self.getlastdevid(self.partnumber,self.sernumber, intcalib)
    else:
      rc= self.lastdevid
    return rc
  def sfp_read(self, slaveadr, htxt, i1, i2):
    self.setmux()
    #print "%s: %s (regs:%d..%d only if !=0):"%(self.devname,htxt, i1,i2)
    line= "" ; einline= 0
    #vals= self.b.i2crm(slaveadr, 0, 128)
    #return vals
    vals=[]    # read byte by byte
    for i in range(i1,i2+1):
      v= self.b.i2crm(slaveadr,i)
      if v==None: break
      vals.append(v[0])
      if v[0] != 0:
        #print "%3d: 0x%2.2x"%(i, v[0])
        line= line+"%3d: 0x%2.2x  "%(i, v[0])
        einline += 1
        if einline>=10: 
          #print line
          line= "" ; einline= 0
    #if einline>0: print line
    return vals 
  def getid_ttcpon(self, prt="any"):
    import math
    if not self.b.seti2cttcmux("ipbus"): return
    self.setmux()
    I2C_FMC_SFPA0_SLAVE_ADDRESS    = 0x50
    I2C_FMC_SFPA2_SLAVE_ADDRESS    = 0x51
    logger_pon= ponlogger()
    slv_addr  = I2C_FMC_SFPA0_SLAVE_ADDRESS

    logger_pon.info('---------------------------------------------------------')
    logger_pon.info('Transceiver diagnostic monitoring')

    # Diagnostic monitoring type
    reg_addr = 92
    reg_value = self.b.i2c_read(slv_addr, reg_addr, 1)
    reg_value = reg_value[0]
    if(reg_value==-1):
        logger_pon.warn('Failed to read from SFP')
        return 0

    # Check implementation
    # Digital diagnostic monitoring implemented?
    if(reg_value & 0b01000000):
        logger_pon.info('SFP digital diagnostic monitoring implemented')	
    else:
        logger_pon.warn('SFP digital diagnostic monitoring not implemented, exiting function')	
        return 0

	# Internally or externally calibrated?
    if(reg_value & 0b00100000):
        internally_calibrated = 1
        logger_pon.info('SFP internally calibrated diagnostic interface')    			
    elif(reg_value & 0b00010000):
        internally_calibrated = 0
        logger_pon.warn('Diagnostic monitoring interface is externally calibrated, this function has to be extended to convert values')    	
    else:
        logger_pon.warn('Unrecognized calibration type')    	

    # Rx power is in OMA or Average power?
    if(reg_value & 0b00001000):
        rx_pwr_unit_avg_noma = 1
        rx_pwr_unit_avg_noma_str = '(AVG)'		
    else:
        rx_pwr_unit_avg_noma = 0
        rx_pwr_unit_avg_noma_str = '(OMA)'

    # Read vendor name
    reg_bank = []
    for reg_addr in range(20,36,2):
        reg_value = self.b.i2c_read(slv_addr, reg_addr,2) # double burst-read access
        reg_value1 = reg_value[0]
        if(reg_value1==-1):
            logger_pon.warn('Failed to read from SFP')
            return 0
        reg_value2 = reg_value[1]			
        reg_bank.append(reg_value1)
        reg_bank.append(reg_value2)
    vendor_name = ''.join(chr(i) for i in reg_bank)

    # Read vendor part number
    reg_bank = []
    for reg_addr in range(40,56,2):
        reg_value = self.b.i2c_read(slv_addr, reg_addr,2) # double burst-read access
        reg_value1 = reg_value[0]
        if(reg_value1==-1):
            logger_pon.warn('Failed to read from SFP')
            return 0
        reg_value2 = reg_value[1]			
        reg_bank.append(reg_value1)
        reg_bank.append(reg_value2)
    vendor_part_number = ''.join(chr(i) for i in reg_bank)

    # Read vendor serial number
    reg_bank = []
    for reg_addr in range(68,84,2):
        reg_value = self.b.i2c_read(slv_addr, reg_addr,2) # double burst-read access
        reg_value1 = reg_value[0]
        if(reg_value1==-1):
            logger_pon.warn('Failed to read from SFP')
            return 0
        reg_value2 = reg_value[1]
        reg_bank.append(reg_value1)
        reg_bank.append(reg_value2)
    vendor_serial_number = ''.join(chr(i) for i in reg_bank)

    slv_addr  = I2C_FMC_SFPA2_SLAVE_ADDRESS

    # Read digital diagnostic
    reg_bank = []
    for reg_addr in range(96,106,2):
        reg_value = self.b.i2c_read(slv_addr, reg_addr,2) # double burst-read access
        reg_value1 = reg_value[0]
        if(reg_value1==-1):
            logger_pon.warn('Failed to read from SFP')
            return 0
        reg_value2 = reg_value[1]			
        reg_bank.append(reg_value1)
        reg_bank.append(reg_value2)

    temp    = (reg_bank[0]<<8) + reg_bank[1]
    vcc     = (reg_bank[2]<<8) + reg_bank[3]
    tx_bias = (reg_bank[4]<<8) + reg_bank[5]	
    tx_pwr  = (reg_bank[6]<<8) + reg_bank[7]
    rx_pwr  = (reg_bank[8]<<8) + reg_bank[9]

    # Convert values if internally calibrated:
    # OBS: The following texts are extracted from SFF-8472 standard for more information, read the standard
    if(internally_calibrated):
        # TEMPERATURE:
        # 1) Internally measured transceiver temperature. Represented as a 16 bit signed twos complement value in
        # increments of 1/256 degrees Celsius
        if(temp&0x8000):
            sign = -1
            temp_conv = (temp ^ 0xFFFF) + 1
        else:
            sign = 1
            temp_conv = temp

        temp_conv = sign*temp_conv*(1.0/256.0) # UNIT: degrees Celsius
        temp_unit = 'degC'

        # SUPPLY VOLTAGE:
        # 2) Internally measured transceiver supply voltage. Represented as a 16 bit unsigned integer with the
        # voltage defined as the full 16 bit value (0-65535) with LSB equal to 100 uVolt
        vcc_conv = (vcc*100e-6) # UNIT: V
        vcc_unit = 'V'

		# TX BIAS CURRENT:
        # 3) Measured TX bias current in uA. Represented as a 16 bit unsigned integer with the current defined as the
        # full 16 bit value (0-65535) with LSB equal to 2 uA
        # OBS: FOR OLT THIS FACTOR IS DIFFERENT - see datasheet (4uA)
        if(vendor_part_number=='LTF7222-BC      ' or vendor_part_number=='SOGX6292-PSGB   '):	
            tx_bias_conv = (tx_bias*4e-6)*1e3 # UNIT: mA
        else:
            tx_bias_conv = (tx_bias*2e-6)*1e3 # UNIT: mA		
        tx_bias_unit = 'mA'

		# TX OUTPUT POWER:
        # 4) Measured TX output power in mW. Represented as a 16 bit unsigned integer with the power defined as
        # the full 16 bit value (0-65535) with LSB equal to 0.1 uW
        if(tx_pwr!=0):
            # OBS: FOR OLT THIS FACTOR IS DIFFERENT - see datasheet (0.2uW)
            if(vendor_part_number=='LTF7222-BC      ' or vendor_part_number=='SOGX6292-PSGB   '):			
                tx_pwr_conv = 10*math.log10((tx_pwr * 0.2 * 1e-6)/1e-3) # UNIT: dBm
            else:
                tx_pwr_conv = 10*math.log10((tx_pwr * 0.1 * 1e-6)/1e-3) # UNIT: dBm			
        else:
            tx_pwr_conv = float('-inf')
        tx_pwr_unit = 'dBm'

		# RX RECEIVED OPTICAL POWER:
        # 5) Measured RX received optical power in mW. Value can represent either average received power or OMA
        # depending upon how bit 3 of byte 92 (A0h) is set. Represented as a 16 bit unsigned integer with the
        # power defined as the full 16 bit value (0-65535) with LSB equal to 0.1 uW
        if(rx_pwr!=0):		
            rx_pwr_conv = 10*math.log10((rx_pwr * 0.1 * 1e-6)/1e-3) # UNIT: dBm
        else:
            rx_pwr_conv = float('-inf')
        rx_pwr_unit = 'dBm'
    else:
        temp_conv    = temp
        vcc_conv     = vcc
        tx_bias_conv = tx_bias
        tx_pwr_conv  = tx_pwr
        rx_pwr_conv  = rx_pwr
        temp_unit    = 'au'
        vcc_unit     = 'au'
        tx_bias_unit = 'au'
        tx_pwr_unit  = 'au'
        rx_pwr_unit  = 'au'

    logger_pon.info('---------------------------------------------------------')
    msg_print = ('Vendor Name          : ' + vendor_name)
    logger_pon.info(msg_print)
    msg_print = ('Vendor Part Number   : ' + vendor_part_number)
    logger_pon.info(msg_print)	
    msg_print = ('Vendor Serial Number : ' + vendor_serial_number)
    logger_pon.info(msg_print)
    logger_pon.info('---------------------------------------------------------')	
    logger_pon.info('Diagnostic reading:')
    msg_print = ('Temperature : %5.1f ' + temp_unit) % temp_conv
    logger_pon.info(msg_print)
    msg_print = ('VCC         : %5.1f ' + vcc_unit) % vcc_conv
    logger_pon.info(msg_print)
    msg_print = ('Tx_bias_cur : %5.1f ' + tx_bias_unit) % tx_bias_conv
    logger_pon.info(msg_print)
    msg_print = ('Tx_Pwr      : %5.1f ' + tx_pwr_unit) % tx_pwr_conv
    logger_pon.info(msg_print)
    msg_print = ('Rx_Pwr      : %5.1f ' + rx_pwr_unit + ' ' + rx_pwr_unit_avg_noma_str) % rx_pwr_conv	
    logger_pon.info(msg_print)
    logger_pon.info('---------------------------------------------------------')
    conv_nraw= 1
    if (conv_nraw): return [internally_calibrated, rx_pwr_unit_avg_noma, temp_conv, vcc_conv, tx_bias_conv, tx_pwr_conv, rx_pwr_conv, vendor_name, vendor_part_number, vendor_serial_number]
    else          : return [internally_calibrated, rx_pwr_unit_avg_noma, temp, vcc, tx_bias, tx_pwr, rx_pwr, vendor_name, vendor_part_number, vendor_serial_number]

  def excmd(self,cmd):
    if len(cmd)<=1:
      self.prthelp()
      return
    if cmd[1]=='id':
      detailed= ""
      forced= False
      if len(cmd)>2 and (cmd[2]=='a'): 
        detailed= "all"
        forced= True
      if len(cmd)>2 and (cmd[2]=='temp'): 
        detailed= "temp"
        forced= True
      # i2c wishbone way:
      sfpid= self.getid(detailed, forced)
      print self.devname,": ",sfpid
    elif cmd[1]=='idttc':
      # ublaze/rev010 way:
      self.getid_ttcpon()
      #self.seti2cttcmux("ipbus")
    elif cmd[1]=='temp':
      t= self.gettemp()
      if t!= None:
        print "%s emperature: %.2f"%(self.devname,self.gettemp())
      else:
        print "%s temp. not povided"%self.devname
    elif cmd[1]=='mod':
      if (self.purpose==None) or type(self.purpose)==types.StringType:
        print "GBT not connected to the SFP"
      else:
        self.purpose.excmd(cmd[1:])
    else:
      self.prthelp()
  def prthelp(self):
    print """ *** I2C connection:
dev SFPX id      -read id (format: type/serial_number Temp Volts OutPwr InPwr)
         id a    -print detailed info in addition to stdout
         id temp -print also temperature (if available)
         idttc      -use Eduardo's code with ipbus
         idttcttc   -use Eduardo's code with i2c over ttcpon NOT AVAILABLE YET

 *** module specific(GBT/OLT) help/commands:
dev SFPX mod [specific_cmd]   
"""
class M24C0x(GenDev):   #
  """ x:1 128 bytes   x:2: 256 bytes eeprom memory
Seems access is ok with (tested on kcu105+ttcpon):
i2cw 0x74 1
i2cwm 0x50 0 0xaa
i2cwm 0x50 1 0xcc
i2cwm 0x50 3 0xee
i2crm 0x50 0 8
i2crm[80:0] ['0xaa', '0xcc', '0x1', '0xee', '0xff', '0xff', '0xff', '0xff']
proto_2 (without fmc card!)
i2cw 0x76 1
i2crm 0x50 0 4
i2crm[80:0] ['0x3', '0x4', '0x0', '0x0']
- cannot be overwritten (no wonder, fmc is not connected)! Why i2c does not report problem?
"""
  def __init__(self, conn, devname="eeprom", addr= 0x50, mux=None):
    GenDev.__init__(self, conn, devname,addr, mux)
  def getid(self):   #
    self.setmux()
    v= self.b.i2crm(self.addr, 0, 4, elog=False)
    if v==None: return None
    return "first 4 bytes: %s"%str(map(hex,v))
  def prthelp(self):
    print """
id               -show first 4 bytes
r from n         -read n bytes from address from
w from b1 b2 b3  -write b1,... from address from
"""
  def myexcmd(self,cmd):
    mad= 0
    n= 4
    if cmd[1]=='id':
      print self.getid()
    elif cmd[1]=='r':
      self.setmux()
      if len(cmd)>2: mad= eval(cmd[2])
      if len(cmd)>3: n= eval(cmd[3])
      v= self.b.i2crm(self.addr, mad, n)
      if v!=None and len(v)>0:
        hexdump.hd_dump(v, 0, len(v)-1, mad)
    elif cmd[1]=='w':
      self.setmux()
      if len(cmd)<4:
        print "dev eeprom w from byte1 byte2 ..."
        return
      mad= eval(cmd[2])
      v=[]
      for ix in range(len(cmd[3:])):
        v.append(eval(cmd[ix+3]))
      print "writing ", v, " from:",mad
      self.b.i2cwm(self.addr, mad, v)
    else:
      self.prthelp()
#
#=== from here kcu105 only
class Si570(GenDev):
  def __init__(self, conn, devname, addr= 0x55, mux=None):
    GenDev.__init__(self, conn, devname,addr, mux)
  def getid(self):
    self.setmux()
    regs= self.b.i2crm(self.addr, 7, 6, elog=False)
    if len(regs)<6: return None
    return "regs[7:13]: %s"%str(map(hex,regs))
  def load(self):
    self.setmux()
    self.b.i2cwm(self.addr, 0x89, 0x10)   # freeze DCO
    rc=self.b.i2cwm(self.addr, 0x07, (0xe0,0x42,0xe3,0x35,0xa2,0xba)) # 240mhz
    # also following, but not any data allowed (0x22 42 bc 2 3d 21 reading
    # when bad values written)
    #self.b.i2cwm(self.addr, 0x07, 0xe0)
    #self.b.i2cwm(self.addr, 0x08, 0x42)
    #self.b.i2cwm(self.addr, 0x09, 0xe3)
    #self.b.i2cwm(self.addr, 0x0a, 0x35)
    #self.b.i2cwm(self.addr, 0x0b, 0xa2)
    #self.b.i2cwm(self.addr, 0x0c, 0xba)
    self.b.i2cwm(self.addr, 0x89, 0x00)
    self.b.i2cwm(self.addr, 0x87, 0x40)   # NewFreq bit (within 10ms)
    #print "bytes written:%d %d"%(rc)
  def read(self):
    self.setmux()
    regs= self.b.i2crm(self.addr, 7, 6)
    print "reg7..12:", map(hex,regs) 
  def myexcmd(self,cmd):
    if cmd[1]=='id':
      self.getid()
    elif cmd[1]=='load':
      self.load()
    elif cmd[1]=='read':
      self.read()
    else:
      self.prthelp()
  def prthelp(self):
    print """
si570 id  -getid
      load
      read
"""
class LEDSttcponfmc(GenDev):
  def __init__(self, conn, devname="leds", addr= 0x38, mux=(0x74,1<<4)):
    GenDev.__init__(self, conn, devname,addr, mux)
  def getid(self):
    self.setmux()
    v= (~self.b.i2cr(self.addr)) & 0xff
    return "ON:0x%x"%v
  def setleds(self, pat):
    self.setmux()
    v= (~pat) & 0xff
    print "Writing 0x%x to LEDSttcponfmc register..."%v
    self.b.i2cw(self.addr, v)
  def prthelp(self):
    print """
id         -get curret pattern
set pat    -set pattern 'pat'
"""
  def myexcmd(self,cmd):
    if cmd[1]=='id':
      print self.getid()
    elif cmd[1]=='set':
      self.setleds(eval(cmd[2]))
    else:
      self.prthelp()
#
#=== from here ctpltu only
class TC74(GenDev):   # ctpltu only?
  def __init__(self, conn, devname="tc74", addr= 0x4d, mux=None):
    GenDev.__init__(self, conn, devname,addr, mux)
  #def setmux(self):       bad idea (b is an instance of I2c...)
  #  GenDev.setmux(self)
  #  self.b.Write("i2cmainttc.mux", 11)
  def getid(self):   # p.79
    self.setmux()
    v= self.b.i2crm(self.addr, 0, 2)   
    # Note: 2 quick readings not ok (i.e. should sleep between)
    if v==None: return None
    return "on-board temp:%d cfg:0x%x"%(v[0],v[1])
  def gettemp(self):
    self.setmux()
    v= self.b.i2crm(self.addr, 0, 2)   
    # Note: 2 quick readings not ok (i.e. should sleep between)
    if v==None: return None
    return v[0]
  def prthelp(self):
    print """
id    -get id
temp  -read temperature
"""
  def myexcmd(self,cmd):
    if cmd[1]=='id':
      print self.devname,": ",self.getid()
    elif cmd[1]=='temp':
      print self.getid()
      #v= self.b.i2crm(self.addr, 0, 2)
      #print "temperature:", v[0],"C", " cfg:0x%x"%v[1]
    else:
      self.prthelp()

