#!/usr/bin/env python
import time,os,string,types,json
def round_floats(o):
  # https://stackoverflow.com/questions/1447287/format-floats-with-standard-json-module
  if isinstance(o, float): return round(o, 2)
  if isinstance(o, dict): return {k: round_floats(v) for k, v in o.items()}
  if isinstance(o, (list, tuple)): return [round_floats(x) for x in o]
  return o
class MetrDB:
  smoothperc= 0.03
  def __init__(self,name):
    self.name= name   #string.split(name,".")[0]
    self.lastdata= {}   # last measured values (even in diffrent times)
    # they may not be written in DB, i.e. it can happen even bigger
    # change is not registered in DB if it changed in small steps
    # todo: introduce lastdatawritten and use it instead when
    # compared for smoothing
    wrkdir= os.getenv("CTP3_WORK_DIRECTORY")
    dbname= os.path.join(wrkdir,self.name+".mdb")
    # Assuming the metrics are written by only 1 process (1 ltud/net)
    self.mdbf= open(dbname, "a")   # "a+" to allow simultaneous add/read
    #
    # following upper (in ttcponnet.py)
    #self.wsup= tri_ws.WSclient(self.name) ; wser= "ok"
    #if self.wsup.ws==None:
    #  wser= "failed"
    print "MetrDB: %s open."%(dbname)
  def Stop(self):
    for ldk in self.lastdata.keys():
      if self.lastdata[ldk]==None:
        del self.lastdata[ldk]
      else:
        self.lastdata[ldk]= None
    ts= time.time()
    #self.mdbf.write("%f %s\n"%(ts, str(self.lastdata)))
    self.myjson(ts, self.lastdata)
    #self.mdbf.write("%f %s\n"%(ts, str(None)))
    self.myjson(ts, None)
    self.mdbf.close()
    #self.wsup.ws_close()
    print 'Stop', self.name
  def myjson(self, ts, dic):
    #self.mdbf.write("%.3f %s\n"%(ts, json.dumps(dic)))
    #self.mdbf.write("%.3f %s\n"%(ts, json.dumps(round_floats(dic))))   # less compact
    self.mdbf.write("%.3f %s\n"%(ts, json.dumps(round_floats(dic),separators=(',',':'))))
  def differs(self, lval, nval):
    if type(lval) != type(nval):
      rc= True
    elif lval==nval:
      rc= False
    elif (type(lval)==types.FloatType) or (type(lval)==types.IntType):
      perc= abs(lval-nval)/((nval+lval)/2.)
      if perc < self.smoothperc:
        #print "metrdb float smoothen %.3f"%perc,lval,nval
        rc= False
      else:
        rc= True
    else:
      rc= True
    return rc
  def Update(self, ts, data):
    """ Unconditional update (no smoothing). Used only for config data.
"""
    #for nkey in self.data.keys():
    #  self.lastdata[nkey]= self.data[nkey]
    self.myjson(ts, data)
    self.mdbf.flush()
  def UpdateSmooth(self, ts, data):
    """self.lastdata should be updated ONLY if different w.r.t. previous value!
If updated always, the data written to DB can drift out of limit, in case
they change in one direction only
"""
    rc= None
    nkd = dict.fromkeys(data.keys(), None)
    dbrec= {}
    # 3 cases when considering previous/new data:
    # last   new
    #A 1     0      0: key not available or value is None
    #B 1     1
    #C 0     1
    for lkey in self.lastdata.keys():
      lval= self.lastdata[lkey]
      nval= None
      if data.has_key(lkey):   #B: last and current available
        del nkd[lkey]
        nval= data[lkey]
        if self.differs(lval, nval):
          dbrec[lkey]= nval
          self.lastdata[lkey]= nval
          #print "metrdb:%s"%lkey,lval,nval
        #else:
          #print "metrdb smoothen (<%.3f) %s:"%(self.smoothperc, lkey),lval,nval
      else:                    #A: lkey has just became None, or 2nd time None
        #if self.differs(lval, nval):
        if lval!= None:
          dbrec[lkey]= nval   # just became None (nval is None)
          self.lastdata[lkey]= nval
      #self.lastdata[lkey]= nval   # data can drift out of the limit
    # nkd: keys for new data only (i.e. not available in prev. measurement)
    for nkey in nkd.keys():    #C: new keys (not available before)
      dbrec[nkey]=self.lastdata[nkey]= data[nkey]
    # prepare message for awm: check olt?_*_dsmon (? olt number (1..) * ONU addr
    # self.wsup.ws_send(awm_msg)
    # update on disk
    if dbrec!={}:   # at least one value changed
      #self.mdbf.write("%f %s\n"%(ts, str(dbrec)))
      try:
        self.myjson(ts, dbrec)
        self.mdbf.flush()
      except:   # IOError:
        import sys
        rc= "Error "+ str(ts) + ":"+str(sys.exc_info())
    return rc
    #print ts,dbrec
 
    #print 'Update done for', self.name
