#!/usr/bin/env python
# needs: pip install websocket-client
from __future__ import print_function
import os,json,time
from websocket import create_connection
class WSclient:
  """ 2 types of connection:
ltud ---wsup---> awmtors <---ws---> web_browser
ltud_server.py(metrdb.py) ----wsup---> awmtors.py <---ws---> web_browser

wsup: one way
ws:   bidirectional
"""
  def __init__(self, netname, contype="wsup"):
    host= os.uname()[1]
    self.wspath= "ws://{0}:8888/{1}/{2}".format(host,contype,netname)
    self.contype= contype
    self.mid="{0}/{1}".format(contype,netname)
    self.ws= None
    self.ws_open()
  def ws_open(self):
    self.nsir= 0   # not sent in row
    rc= 0; msg= "ok"
    if self.ws==None:
      try:
        self.ws = create_connection(self.wspath)
      except:
        msg= "Error: cannot open wspath: " + self.wspath
        self.ws= None
    else:
      msg= "Warning: ws already open wspath:"+ self.wspath
    #print(msg)
    return(rc,msg)
  def ws_close(self):
    rc= 0; msg= "ok"
    if self.ws:
      self.ws.close()
      self.ws= None
    else:
      msg="{0} was closed already".format(self.wspath)
    return(rc,msg)
  def ws_send(self, msgo):
    #if self.ws.connected:   # always True (even with broken connection)
    if self.ws:
      rc= self.ws.send(json.dumps(msgo))
      print("{0}.send rc:{1} {2}".format(self.mid,rc, msgo))
      if self.contype=="wsup":
        greeting = self.ws.recv()
        print("< %s"%greeting)
      self.nsir= 0
    else:
      self.nsir= self.nsir+1
      if (self.nsir < 4) or ((self.nsir % 100)== 0):
        print("ws_send to {0}: not sent (not connected)".format(self.mid))
class AWMclient(WSclient):
  def __init__(self, netname, contype="wsup"):
    WSclient.__init__(self, netname, contype)
    self.allonus= {}   # "olt?_*_dsmon":[onuN, last_dsmon_value, state]
    # ?: 1...     *: onuAddr
    # onuN: 1... 
    # last_dsmon_value: [fecS,fecD,sc_crc_detected]
    # state: 0 (on), 1(err), 2(not known)
    self.changedonus=[]  # [(oltN,onuN,2/1/0),...]
    self.dbgchonu= False   # True: dbg mode
  def update_cfg(self, oltname, onuaddr, ixonu1):
    self.allonus[oltname+"_"+onuaddr+"_dsmon"]= [ixonu1,[0,0,0], 2]
  def update(self, keyname, dsmon_v):
    newstate= self.allonus[keyname][2]
    if self.dbgchonu == True:
      # in dbg mode: state= state+1
      msg= "dbg {0}: {1}-> ".format(keyname, str(newstate))
      if newstate==2:
        newstate= 0
      else:
        newstate= newstate+1
    else:
      msg= "{0}: {1}-> ".format(keyname, str(newstate))
      #print("AWMclient update...")
      if dsmon_v==None:
        newstate= 2   # unknown
      else:
        if self.allonus[keyname][1]!=dsmon_v:
          newstate= 1    # err
        else:
          newstate= 0    # ok
    msg= msg+str(newstate)
    self.allonus[keyname][1]= dsmon_v
    if newstate != self.allonus[keyname][2]:
      oltn= int(keyname[3])
      self.changedonus.append((oltn,self.allonus[keyname][0],newstate))
      self.allonus[keyname][2]= newstate
    return(0, msg)
  def sendawm(self):
    timestamp= time.time()
    if len(self.changedonus)>0:
      #send 
      msgo= {"excmd":"", "onus":self.changedonus, "epochtime":timestamp}
      self.ws_send(msgo)
      self.changedonus=[]
    else:
      self.ws_send({"epochtime":timestamp})
  def forceupdate(self):
    for onuk in self.allonus:
      self.allonus[onuk][2]= 2   # unknown for all
    return(0,"ok")
  def set_dbgchonu(self):
    if self.dbgchonu == True:
      self.dbgchonu= False
    else:
      self.dbgchonu= True
    return(0,"ok dbgchonu:{0}".format(str(self.dbgchonu)))
